# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stochastic Gradient Descent (Adam) Approx. Nash Solver w/ Annealing."""

from absl import logging  # pylint:disable=unused-import

import jax
import jax.numpy as jnp

import numpy as np

import optax

from scipy import special

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import exploitability as exp


class Solver(object):
  """Adam Solver with temperature annealing."""

  def __init__(self, temperature=1., proj_grad=True, lrs=(1e-2, 1e-1),
               exp_thresh=-1., rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    if temperature < 0.:
      raise ValueError("temperature must be non-negative")
    self.num_players = None
    self.temperature = temperature
    self.proj_grad = proj_grad
    self.rnd_init = rnd_init
    self.lrs = lrs
    self.num_estimates = 2
    self.exp_thresh = exp_thresh
    self.has_aux = True
    self.aux_errors = []

    self.update = self.descent_step

    self.opt = optax.adam(learning_rate=lrs[0])
    self.opt_state = self.opt.init(jnp.zeros(1))

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if self.rnd_init:
      init_dist = self.random.rand(num_strats)
    else:
      init_dist = np.ones(num_strats)
    init_dist /= init_dist.sum()
    init_y = np.zeros(num_strats)
    init_anneal_steps = 0

    init_params = jnp.array(dist_to_logits(init_dist))

    self.opt_state = self.opt.init(init_params)

    return (init_dist, init_y, init_anneal_steps)

  def record_aux_errors(self, grads):
    """Record errors for the auxiliary variables."""
    grad_y = grads[1]
    self.aux_errors.append([np.linalg.norm(grad_y)])

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients (and exploitabilities) for all parameters.

    Args:
      params: tuple of params (dist, y, anneal_steps), see gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      tuple of gradients (grad_dist, grad_y, grad_anneal_steps), see gradients
      unregularized exploitability (stochastic estimate)
      shannon entropy regularized exploitability (stochastic estimate)
    """
    return self.gradients(*params, payoff_matrices, self.num_players,
                          self.temperature, self.proj_grad)

  def exploitability(self, params, payoff_matrices):
    """Compute and return shannon entropy regularized exploitability.

    Args:
      params: tuple of params (dist, y), see qre.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability of current dist
    """
    return exp.qre_exploitability(params, payoff_matrices, self.temperature)

  def gradients(self, dist: np.ndarray, y: np.ndarray, anneal_steps: int,
                payoff_matrices, num_players,
                temperature=0., proj_grad=True
                ) -> tuple[tuple[np.ndarray, np.ndarray, int], float, float]:
    """Computes exploitablity gradient and aux variable gradients.

    Args:
      dist: 1-d np.array, current estimate of nash distribution
      y: 1-d np.array (same shape as dist), current estimate of payoff gradient
      anneal_steps: int, elapsed num steps since last anneal
      payoff_matrices: 2 (>=2 x A x A) np.arrays, payoffs for each joint action
      num_players: int, number of players, in case payoff_matrices is
        abbreviated
      temperature: non-negative float, default 0.
      proj_grad: bool, if True, projects dist gradient onto simplex
    Returns:
      gradient of exploitability w.r.t. (dist, anneal_steps) as tuple
      unregularized exploitability (stochastic estimate)
      shannon entropy regularized exploitability (stochastic estimate)
    """

    grad_dist = loss_gradients(dist, payoff_matrices, num_players, temperature,
                               proj_grad)[0][0]

    a = 0  # 2 samples (a, b) needed for unbiased estimation
    p_0 = 0  # player 0 index
    nabla = payoff_matrices[a][p_0].dot(dist)
    grad_y = y - nabla

    if temperature >= 1e-3:
      br = special.softmax(y / temperature)
    else:
      power = np.inf
      s = np.linalg.norm(y, ord=power)
      br = np.zeros_like(dist)
      maxima = (y == s)
      br[maxima] = 1. / maxima.sum()

    unreg_exp = np.max(y) - y.dot(dist)
    entr_br = temperature * special.entr(br).sum()
    entr_dist = temperature * special.entr(dist).sum()
    reg_exp = y.dot(br - dist) + entr_br - entr_dist

    if reg_exp < self.exp_thresh:
      self.temperature = np.clip(temperature / 2., 0., np.inf)
      grad_anneal_steps = -anneal_steps
    else:
      grad_anneal_steps = 1

    return (grad_dist, grad_y, grad_anneal_steps), unreg_exp, reg_exp

  def descent_step(self, params, grads, t, eps=0.):
    """Gradient descent on exploitability wrt logits.

    Args:
      params: tuple of variables to be updated (dist, y, anneal_steps)
      grads: tuple of variable gradients (grad_dist, grad_y, grad_anneal_steps)
      t: int, solver iteration
      eps: float > 0, force all probabilities >= eps / dim(dist) (unused)
    Returns:
      new_params: tuple of update params (new_dist, new_y, new_anneal_steps)
    """
    del eps

    dist = params[0]
    grads_dist = grads[0]

    dist_jnp = jnp.array(dist)
    grads_dist_jnp = jnp.array(grads_dist)

    # map dist to logits and grads to grad_logits using jacobian
    logits = dist_to_logits(dist)
    grads_logits = jax.jvp(dist_to_logits, [dist_jnp], [grads_dist_jnp])[1]

    opt_updates, self.opt_state = self.opt.update(grads_logits,
                                                  self.opt_state,
                                                  logits)

    new_logits = optax.apply_updates(logits, opt_updates)

    new_dist = logits_to_dist(new_logits)
    new_dist = np.array(new_dist)

    lr_y = self.lrs[1]
    lr_y = np.clip(1 / float(t + 1), lr_y, np.inf)
    new_y = params[1] - lr_y * grads[1]

    new_anneal_steps = params[2] + grads[2]

    return (new_dist, new_y, new_anneal_steps)


def logits_to_dist(logits):
  logits_ext = jnp.append(logits, 0.)
  payoff = jax.nn.softmax(logits_ext)
  return payoff


def dist_to_logits(dist, eps=1e-8):
  # dist[-1] = exp(logits[-1]) / Z = exp(0) / Z
  z = 1 / jnp.clip(dist[-1], eps, 1.)
  logits = jnp.log(jnp.clip(dist[:-1] * z, eps, np.inf))
  return logits


def loss_gradients(dist, payoff_matrices, num_players, temperature=0.,
                   proj_grad=True):
  """Computes exploitablity gradient.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_matrices: 2 (>=2 x A x A) np.arrays, payoffs for each joint action
    num_players: int, number of players, in case payoff_matrices is abbreviated
    temperature: non-negative float, default 0.
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist) as tuple
    unregularized exploitability (stochastic estimate)
    shannon regularized exploitability (stochastic estimate)
  """
  del num_players
  # if consulting paper https://arxiv.org/abs/2310.06689, code assumes eta = 1
  tau = temperature

  a, b = 0, 1  # 2 samples needed for unbiased estimation
  p_0, p_1 = 0, 1  # player 0 index, player 1 index
  hess_0_01_a = payoff_matrices[a][p_0]
  hess_1_01_a = payoff_matrices[a][p_1]
  hess_0_01_b = payoff_matrices[b][p_0]

  pg_0_a = simplex.project_grad(hess_0_01_a.dot(dist))
  pg_0_b = simplex.project_grad(hess_0_01_b.dot(dist))

  unreg_exp = np.dot(pg_0_a, pg_0_b)

  if tau > 0.:
    log_dist_safe = np.clip(np.log(dist), -40, 0)
    entr_grad_proj = simplex.project_grad(-tau * (log_dist_safe + 1))
  else:
    entr_grad_proj = 0.
  pg_0_a_entr = pg_0_a + entr_grad_proj
  pg_0_b_entr = pg_0_b + entr_grad_proj
  pg_0_entr = 0.5 * (pg_0_a_entr + pg_0_b_entr)
  pg_1_b_entr = pg_0_b_entr

  reg_exp = np.dot(pg_0_a_entr, pg_0_b_entr)

  # then construct unbiased stochastic gradient
  grad_dist = 2. * hess_1_01_a.dot(pg_1_b_entr)
  if tau > 0.:
    grad_dist += 2. * -tau * pg_0_entr / dist

  if proj_grad:
    grad_dist = simplex.project_grad(grad_dist)

  return (grad_dist,), unreg_exp, reg_exp



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.jpsro."""

import itertools

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.algorithms import jpsro
import pyspiel

GAMES = (
    "sheriff_2p_gabriele",
)
SWEEP_KWARGS = [
    dict(  # pylint: disable=g-complex-comprehension
        game_name=game,
        iterations=iterations,
        policy_init=policy_init,
        update_players_strategy=update_players_strategy,
        target_equilibrium=target_equilibrium,
        br_selection=br_selection,
        train_meta_solver=train_meta_solver,
        eval_meta_solver=eval_meta_solver,
        ignore_repeats=ignore_repeats,
    ) for (
        iterations,
        game,
        policy_init,
        update_players_strategy,
        target_equilibrium,
        br_selection,
        train_meta_solver,
        eval_meta_solver,
        ignore_repeats) in itertools.product(
            [2],
            GAMES,
            jpsro.INIT_POLICIES,
            jpsro.UPDATE_PLAYERS_STRATEGY,
            jpsro.BRS,
            jpsro.BR_SELECTIONS,
            jpsro.META_SOLVERS,
            ["mwcce"],
            [True, False])
]
TEST_COUNT_LIMIT = 100

interval = len(SWEEP_KWARGS) // TEST_COUNT_LIMIT
interval = interval if interval % 2 != 0 else interval + 1  # Odd interval.
SWEEP_KWARGS = SWEEP_KWARGS[::interval]


def get_game(game_name):
  """Returns the game."""
  if game_name == "kuhn_poker_3p":
    game_name = "kuhn_poker"
    game_kwargs = {"players": int(3)}
  elif game_name == "trade_comm_2p_2i":
    game_name = "trade_comm"
    game_kwargs = {"num_items": int(2)}
  elif game_name == "sheriff_2p_gabriele":
    game_name = "sheriff"
    game_kwargs = {
        "item_penalty": float(1.0),
        "item_value": float(5.0),
        "max_bribe": int(2),
        "max_items": int(10),
        "num_rounds": int(2),
        "sheriff_penalty": float(1.0),
    }

  else:
    raise ValueError("Unrecognised game: %s" % game_name)
  return pyspiel.load_game_as_turn_based(game_name, game_kwargs)


class JPSROTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(*SWEEP_KWARGS)
  def test_jpsro_cce(self, **kwargs):
    game = get_game(kwargs["game_name"])
    jpsro.run_loop(game=game, **kwargs)


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mean Field Game on periodic domain with aversion cost.

This is a demonstration of implementing a mean field game in Python. The model
is an approximation of a continuous space, continuous time model introduced
to study ergodic MFG with explicit solution in:
Almulla, N.; Ferreira, R.; and Gomes, D. 2017.
Two numerical approaches to stationary mean-field games. Dyn. Games Appl.
7(4):657-682.

See also:
Elie, R., Perolat, J., Laurière, M., Geist, M., & Pietquin, O. (2020, April).
On the convergence of model free learning in mean field games.
In Proceedings of the AAAI Conference on Artificial Intelligence
(Vol. 34, No. 05, pp. 7143-7150).
"""

import functools
import math
from typing import Any, List, Mapping

import numpy as np
import scipy.stats

from open_spiel.python import observation
import pyspiel

_NUM_PLAYERS = 1
_SIZE = 21
_HORIZON = 20
_VOLATILITY = 1.0
_COEF_AVERSION = 1.0
_DELTA_T = 0.01
_X_MIN = 0.0
_X_MAX = 1.0
_N_ACTIONS_PER_SIDE = 10

_DEFAULT_PARAMS = {
    "size": _SIZE,
    "horizon": _HORIZON,
    "dt": _DELTA_T,
    "xmin": _X_MIN,
    "xmax": _X_MAX,
    "n_actions_per_side": _N_ACTIONS_PER_SIDE,
    "volatility": _VOLATILITY,
    "coef_aversion": _COEF_AVERSION,
}

_GAME_TYPE = pyspiel.GameType(
    short_name="python_mfg_periodic_aversion",
    long_name="Mean-Field Periodic Aversion Game",
    dynamics=pyspiel.GameType.Dynamics.MEAN_FIELD,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.GENERAL_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=False,
    provides_information_state_tensor=False,
    provides_observation_string=True,
    provides_observation_tensor=True,
    parameter_specification=_DEFAULT_PARAMS,
)


@functools.lru_cache(maxsize=None)
def _state_to_str(x, t, player_id):
  """A string that uniquely identifies (x, t, player_id)."""
  if int(player_id) == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
    return f"(t={t}, pos={x})"
  if player_id == pyspiel.PlayerId.MEAN_FIELD:
    return f"(t={t}_a, pos={x})"
  if player_id == pyspiel.PlayerId.CHANCE:
    return f"(t={t}_a_mu, pos={x})"


class MFGPeriodicAversionGame(pyspiel.Game):
  """A Mean-Field Game on periodic domain with crowd aversion cost.

  A game starts by an initial chance node that select the initial state
  of the player in the MFG.
  Then the game sequentially alternates between:
    - An action selection node (where the player id is >= 0)
    - A chance node (the player id is pyspiel.PlayerId.CHANCE)
    - A Mean Field node (the player id is pyspiel.PlayerId.MEAN_FIELD)
  """

  # pylint:disable=dangerous-default-value
  def __init__(self, params: Mapping[str, Any] = _DEFAULT_PARAMS):
    self.size = params.get("size", _SIZE)  # number of states
    self.horizon = params.get("horizon", _HORIZON)  # number of time steps
    self.dt = params.get("dt", _DELTA_T)  # size of one step in time
    self.xmin = params.get("xmin", _X_MIN)  # smallest position
    self.xmax = params.get("xmax", _X_MAX)  # largest position
    self.dx = (self.xmax - self.xmin) / (
        self.size - 1
    )  # size of one step in space
    self.n_actions_per_side = params.get(
        "n_actions_per_side", _N_ACTIONS_PER_SIDE
    )  # number of actions on each side, for both players and noise
    self.volatility = params.get("volatility", _VOLATILITY)
    self.coef_aversion = params.get("coef_aversion", _COEF_AVERSION)

    game_info = pyspiel.GameInfo(
        num_distinct_actions=2 * self.n_actions_per_side + 1,
        max_chance_outcomes=2 * self.n_actions_per_side + 1,
        num_players=_NUM_PLAYERS,
        min_utility=-np.inf,
        max_utility=+np.inf,
        utility_sum=0.0,
        max_game_length=self.horizon,
    )
    super().__init__(_GAME_TYPE, game_info, params)

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return MFGPeriodicAversionState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    if (iig_obs_type is None) or (
        iig_obs_type.public_info and not iig_obs_type.perfect_recall
    ):
      return Observer(params, self)
    return observation.IIGObserverForPublicInfoGame(iig_obs_type, params)

  def max_chance_nodes_in_history(self):
    """Maximun chance nodes in game history."""
    return self.horizon + 1


class MFGPeriodicAversionState(pyspiel.State):
  """A Mean Field Normal-Form state.

  In this class, x and action are integers. They are converted, when needed, to
  spatial variables by using a scaling factor representing the size of a step in
  space and by shifting them depending on the minimal allowed value.
  """

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    # Initial state where the initial position is chosen according to
    # an initial distribution.
    self._player_id = pyspiel.PlayerId.CHANCE

    self._last_action = game.n_actions_per_side  # neutral action
    self.tick = 0
    self.x = None
    self.return_value = 0.0

    self.game = game

    self.size = game.size
    self.horizon = game.horizon
    self.dt = game.dt
    self.xmin = game.xmin
    self.xmax = game.xmax
    self.dx = game.dx
    self.da = game.dx
    self.n_actions_per_side = game.n_actions_per_side
    self.volatility = game.volatility
    self.coef_aversion = game.coef_aversion

    # Represents the current probability distribution over game states.
    # Initialized with a uniform distribution.
    self._distribution = [1.0 / self.size for _ in range(self.size)]

  def to_string(self):
    return self.state_to_str(self.x, self.tick)

  def state_to_str(self, x, tick, player_id=pyspiel.PlayerId.DEFAULT_PLAYER_ID):
    """A string that uniquely identify a triplet x, t, player_id."""
    if self.x is None:
      return "initial"
    if self._player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      return "({}, {})".format(x, tick)
    elif self._player_id == pyspiel.PlayerId.MEAN_FIELD:
      return "({}, {})_a".format(x, tick)
    elif self._player_id == pyspiel.PlayerId.CHANCE:
      return "({}, {})_a_mu".format(x, tick)
    raise ValueError(
        "player_id is not mean field, chance or default player id."
    )

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every perfect-information sequential-move game.

  @property
  def n_actions(self):
    return 2 * self.n_actions_per_side + 1

  def _legal_actions(self, player):
    """Returns a list of legal actions for player and MFG nodes."""
    if player == pyspiel.PlayerId.MEAN_FIELD:
      return []
    if (
        player == pyspiel.PlayerId.DEFAULT_PLAYER_ID
        and player == self.current_player()
    ):
      return list(range(self.n_actions))
    raise ValueError(
        f"Unexpected player {player}. "
        "Expected a mean field or current player 0."
    )

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    if self._player_id == pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "_apply_action should not be called at a MEAN_FIELD state."
      )
    self.return_value = self._rewards()

    assert (
        self._player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID
        or self._player_id == pyspiel.PlayerId.CHANCE
    )

    if self.x is None:
      self.x = action
      self._player_id = pyspiel.PlayerId.DEFAULT_PLAYER_ID
      return

    if action < 0 or action >= self.n_actions:
      raise ValueError(
          "The action is between 0 and {} at any node".format(self.n_actions)
      )

    self.x = (self.x + action - self.n_actions_per_side) % self.size
    if self._player_id == pyspiel.PlayerId.CHANCE:
      self._player_id = pyspiel.PlayerId.MEAN_FIELD
      self.tick += 1
    elif self._player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      self._last_action = action
      self._player_id = pyspiel.PlayerId.CHANCE

  def _action_to_string(self, player, action):
    """Action -> string."""
    del player
    return str(action - self.n_actions_per_side)

  def action_to_move(self, action):
    return (action - self.n_actions_per_side) * self.da

  def state_to_position(self, state):
    return state * self.dx + self.xmin

  def position_to_state(self, position):
    return round((position - self.xmin) / self.dx)

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    if self.x is None:
      # Initial distribution
      return list(enumerate(self._distribution))
    actions = np.array(
        [(a - self.n_actions_per_side) * self.da for a in range(self.n_actions)]
    )
    stddev = self.volatility * math.sqrt(self.dt)
    probas = scipy.stats.norm.pdf(actions, scale=stddev)
    probas /= np.sum(probas)
    return [(act, p) for act, p in zip(list(range(self.n_actions)), probas)]

  def distribution_support(self):
    """return a list of state string."""
    return [
        self.state_to_str(i, self.tick, player_id=pyspiel.PlayerId.MEAN_FIELD)
        for i in range(self.size)
    ]

  def get_state_proba(self, state: int) -> float:
    """Gets the probability of a position in the current distrib.

    Args:
      state: state requested.

    Returns:
      The probability for the provided position.
    """
    assert state >= 0, state
    assert state < self.size, state
    # This logic needs to match the ordering defined in distribution_support().
    index = state
    assert 0 <= index < len(self._distribution), (
        f"Invalid index {index} vs dist length:"
        f" {len(self._distribution)}, state={state},"
        f" state={self}"
    )
    return self._distribution[index]

  def update_distribution(self, distribution):
    """This function is central and specific to the logic of the MFG.

    Args:
      distribution: a distribution to register.  - function should be called
        when the node is in MEAN_FIELD state. - distribution are probabilities
        that correspond to each game state given by distribution_support.
    """
    if self._player_id != pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "update_distribution should only be called at a MEAN_FIELD state."
      )
    self._distribution = distribution.copy()
    self._player_id = pyspiel.PlayerId.DEFAULT_PLAYER_ID

  @property
  def t(self):
    return self.tick

  def is_terminal(self):
    """Returns True if the game is over."""
    return self.t >= self.horizon

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    return self._player_id

  def _rewards(self):
    """Reward for the player for this state."""
    if self._player_id != pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      return 0.0
    assert self.x is not None
    velocity = self.action_to_move(self._last_action) / self.dt
    action_r = -0.5 * velocity**2
    eps = 1e-15
    mu_x = self.get_state_proba(self.x) / self.dx  # represents the density
    # The density should have an integral equal to 1; here sum_x mu_x * dx = 1
    aversion_r = -np.log(mu_x + eps)
    pos = self.state_to_position(self.x)
    pix2 = 2 * np.pi * pos
    geom_r = (
        self.volatility * 2 * np.pi**2 * np.sin(pix2)
        - 2 * np.pi**2 * np.cos(pix2) ** 2
        + (2 / self.volatility**2) * np.sin(pix2)
    )
    return (action_r + self.coef_aversion * aversion_r + geom_r) * self.dt

  def rewards(self) -> List[float]:
    """Rewards for all players."""
    # For now, only single-population (single-player) mean field games
    # are supported.
    return [self._rewards()]

  def _returns(self):
    """Returns is the sum of all payoffs collected so far."""
    return self.return_value + self._rewards()

  def returns(self) -> List[float]:
    """Returns for all players."""
    # For now, only single-population (single-player) mean field games
    # are supported.
    return [self._returns()]

  def __str__(self):
    """A string that uniquely identify the current state."""
    return self.state_to_str(
        x=self.x, tick=self.tick, player_id=self._player_id
    )


class Observer:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, params, game):
    """Initializes an empty observation tensor."""
    del params

    self.size = game.size
    self.horizon = game.horizon
    # +1 to allow t == horizon.
    self.tensor = np.zeros(self.size + self.horizon + 1, np.float32)
    self.dict = {"x": self.tensor[: self.size], "t": self.tensor[self.size :]}

  def set_from(self, state, player: int):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    del player
    # We update the observation via the shaped tensor since indexing is more
    # convenient than with the 1-D tensor. Both are views onto the same memory.
    self.tensor.fill(0)
    # state.x is None for the initial (blank) state, don't set any
    # position bit in that case.
    if state.x is not None:
      if state.x < 0 or state.x > self.size:
        raise ValueError(
            f"Expected {state} positions to be in [0, {self.size})"
        )
      self.dict["x"][state.x] = 1
    if not 0 <= state.tick <= self.horizon:
      raise ValueError(f"Expected {state} time to be in [0, {self.horizon}]")
    self.dict["t"][state.tick] = 1

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    del player
    return state.to_string()


pyspiel.register_game(_GAME_TYPE, MFGPeriodicAversionGame)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to compute Shapley values and their approximations."""

import itertools
import numpy as np
from open_spiel.python.coalitional_games import coalitional_game


def compute_shapley_values(
    game: coalitional_game.CoalitionalGame,
) -> np.ndarray:
  """Compute the Shapley values exactly.

  Uses Eq (2) of Mitchell et al. "Sampling Permutations for Shapley Value
  Estimation". https://people.math.sc.edu/cooper/shapley.pdf

  Args:
    game: the game to compute Shapley values for.

  Returns:
    shapley_values: a numpy array of Shapley values per player.
  """

  shapley_values_sum = np.zeros(game.num_players(), dtype=float)
  coalition = np.zeros(game.num_players(), dtype=int)
  empty_coalition_value = game.coalition_value(coalition)
  num_perms = 0
  for perm_tup in itertools.permutations(range(game.num_players())):
    perm = list(perm_tup)
    value_with = empty_coalition_value
    coalition.fill(0)
    for idx in range(game.num_players()):
      value_without = value_with  # re-use the one computed from the last iter
      i = perm[idx]
      coalition[i] = 1
      value_with = game.coalition_value(coalition)
      shapley_values_sum[i] += value_with - value_without
    num_perms += 1
  return shapley_values_sum / num_perms


def compute_approximate_shapley_values(
    game: coalitional_game.CoalitionalGame,
    num_samples: int,
) -> np.ndarray:
  """Compute the Shapley values using Monte Carlo estimation.

  Specifically, applies the implementation described in Section 2.3 of Mitchell
  et al. "Sampling Permutations for Shapley Value Estimation".
  https://people.math.sc.edu/cooper/shapley.pdf

  Args:
    game: the game to compute Shapley values for.
    num_samples: number of permutations to sample

  Returns:
    shapley_values: a numpy array of Shapley values per player.
  """

  shapley_values_sum = np.zeros(game.num_players(), dtype=float)
  coalition = np.zeros(game.num_players(), dtype=int)
  empty_coalition_value = game.coalition_value(coalition)
  for _ in range(num_samples):
    perm = np.random.permutation(game.num_players())
    value_with = empty_coalition_value
    coalition.fill(0)
    for idx in range(game.num_players()):
      value_without = value_with  # re-use the one computed from the last iter
      i = perm[idx]
      coalition[i] = 1
      value_with = game.coalition_value(coalition)
      shapley_values_sum[i] += value_with - value_without
  return shapley_values_sum / num_samples




# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mirror descent."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import mirror_descent
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.games import crowd_modelling  # pylint: disable=unused-import
import pyspiel


class MirrorDescentTest(parameterized.TestCase):

  @parameterized.named_parameters(('python', 'python_mfg_crowd_modelling'),
                                  ('cpp', 'mfg_crowd_modelling'))
  def test_fp(self, name):
    """Checks if mirror descent works."""
    game = pyspiel.load_game(name)
    md = mirror_descent.MirrorDescent(game, value.TabularValueFunction(game))
    for _ in range(10):
      md.iteration()
    md_policy = md.get_policy()
    nash_conv_md = nash_conv.NashConv(game, md_policy)

    self.assertAlmostEqual(nash_conv_md.nash_conv(), 2.2730324915546056)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python dynamic routing game utils."""

from absl.testing import absltest
from open_spiel.python.games import dynamic_routing_utils as utils


class NetworkTest(absltest.TestCase):
  """Tests for Network class."""

  def setUp(self):
    """Create a network O->A->D for testing."""
    super().setUp()
    self.network = utils.Network({"O": ["A"], "A": ["D"], "D": []})

  def test_adjacency_list_init(self):
    """Test class instanciation with adjacency list."""
    self.assertEqual(self.network.num_links(), 2)
    self.assertEqual(self.network.get_successors("O"), ["A"])
    self.assertEqual(self.network.get_successors("A"), ["D"])
    self.assertEqual(self.network.get_successors("D"), [])
    self.assertTrue(self.network.is_location_at_sink_node("A->D"))
    self.assertFalse(self.network.is_location_at_sink_node("O->A"))
    self.assertEqual(self.network.get_action_id_from_movement("A", "D"), 1)
    self.assertEqual(self.network.get_action_id_from_movement("O", "A"), 2)
    self.assertEqual(self.network.get_road_section_from_action_id(1), "A->D")
    self.assertEqual(self.network.get_road_section_from_action_id(2), "O->A")

  def test_get_successors_with_wrong_node(self):
    """Test get successors on non existing node."""
    with self.assertRaises(KeyError):
      self.network.get_successors("Z")

  def test_get_action_id_without_connected_nodes(self):
    """Test get actions id on non connected nodes."""
    with self.assertRaises(KeyError):
      self.network.get_action_id_from_movement("O", "D")

  def test_get_action_id_with_wrong_nodes(self):
    """Test get actions id on non existing node."""
    with self.assertRaises(KeyError):
      self.network.get_action_id_from_movement("Z", "D")

  def test_is_location_at_sink_noded_with_wrong_road_section(self):
    """Test is_location_at_sink_node on non existing second node."""
    with self.assertRaises(KeyError):
      self.network.is_location_at_sink_node("A->Z")

  def test_is_location_at_sink_noded_with_wrong_road_section_2(self):
    """Test is_location_at_sink_node on non existing first node."""
    with self.assertRaises(KeyError):
      self.network.is_location_at_sink_node("Z->D")

  def test_is_location_at_sink_noded_with_wrong_arg(self):
    """Test is_location_at_sink_node on wrong link str representation."""
    with self.assertRaises(ValueError):
      self.network.is_location_at_sink_node("D")

  def test_get_road_section_with_action_id(self):
    """Test get_road_section_from_action_id on non possible action."""
    with self.assertRaises(KeyError):
      self.network.get_road_section_from_action_id(0)

  def test_num_links_method(self):
    # Write.
    pass

  def test_num_actions_method(self):
    # Write.
    pass

  def test_links(self):
    # Write.
    pass

  def test_check_list_of_vehicles_is_correct_method(self):
    # Write.
    pass

  def test_check_list_of_od_demand_is_correct_method(self):
    # Write.
    pass

  def test_str_method(self):
    # Write.
    pass

  def test_get_travel_time_methods(self):
    # Write.
    pass

  def test_assert_valid_action_methods(self):
    # Write.
    pass

  def test_default_travel_time_methods(self):
    # Write.
    pass

  def test_customable_travel_time_methods(self):
    # Write.
    pass


class VehicleTest(absltest.TestCase):
  """Tests for Vehicle class."""

  def test_vehicle_1(self):
    """Test instanciation of Vehicle."""
    vehicle = utils.Vehicle("O->A", "B->D")
    self.assertEqual(vehicle.destination, "B->D")
    self.assertEqual(vehicle.origin, "O->A")
    self.assertEqual(vehicle.departure_time, 0)

  def test_vehicle_2(self):
    """Test instanciation of with departure time."""
    vehicle = utils.Vehicle("O->A", "B->D", 10.5)
    self.assertEqual(vehicle.origin, "O->A")
    self.assertEqual(vehicle.destination, "B->D")
    self.assertEqual(vehicle.departure_time, 10.5)


class OriginDestinationDemandTest(absltest.TestCase):
  """Tests for OriginDestinationDemand class."""

  def test_od_demand_1(self):
    """Test instanciation of OD demand."""
    od_demand = utils.OriginDestinationDemand("O->A", "B->D", 0, 30)
    self.assertEqual(od_demand.destination, "B->D")
    self.assertEqual(od_demand.origin, "O->A")
    self.assertEqual(od_demand.departure_time, 0)
    self.assertEqual(od_demand.counts, 30)

  def test_od_demand_2(self):
    """Test instanciation of OD demand."""
    od_demand = utils.OriginDestinationDemand("O->A", "B->D", 10.5, 43.2)
    self.assertEqual(od_demand.origin, "O->A")
    self.assertEqual(od_demand.destination, "B->D")
    self.assertEqual(od_demand.departure_time, 10.5)
    self.assertEqual(od_demand.counts, 43.2)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.pybind11.pyspiel."""

from absl.testing import absltest

from open_spiel.python import rl_environment
import pyspiel


class RLEnvironmentTest(absltest.TestCase):

  def test_create_game(self):
    env = rl_environment.Environment("tic_tac_toe")
    self.assertEqual(env.is_turn_based, True)
    self.assertEqual(env.num_players, 2)

  def test_create_game_with_args(self):
    env = rl_environment.Environment("kuhn_poker", **{"players": 3})
    self.assertEqual(env.is_turn_based, True)
    self.assertEqual(env.num_players, 3)

  def test_create_env_from_game_instance(self):
    game = pyspiel.load_game("tic_tac_toe")
    env = rl_environment.Environment(game)
    self.assertEqual(env.is_turn_based, True)
    self.assertEqual(env.num_players, 2)

  def test_reset(self):
    env = rl_environment.Environment("kuhn_poker", **{"players": 3})
    time_step = env.reset()
    self.assertEqual(time_step.observations["current_player"], 0)
    self.assertEmpty(time_step.observations["serialized_state"], 0)
    self.assertLen(time_step.observations["info_state"], 3)
    self.assertLen(time_step.observations["legal_actions"], 3)
    self.assertIsNone(time_step.rewards)
    self.assertIsNone(time_step.discounts)
    self.assertEqual(time_step.step_type.first(), True)

  def test_initial_info_state_is_decision_node(self):
    env = rl_environment.Environment("kuhn_poker")
    time_step = env.reset()
    self.assertEqual(time_step.step_type.first(), True)
    self.assertEqual(env.is_chance_node, False)

  def test_full_game(self):
    env = rl_environment.Environment("tic_tac_toe", include_full_state=True)
    _ = env.reset()
    time_step = env.step([0])
    self.assertEqual(time_step.observations["current_player"], 1)
    self.assertLen(time_step.observations["info_state"], 2)
    self.assertLen(time_step.observations["legal_actions"], 2)
    self.assertLen(time_step.rewards, 2)
    self.assertLen(time_step.discounts, 2)
    self.assertLen(time_step.observations, 4)

    # O X O   # Moves 0, 1, 2
    # X O X   # Moves 3, 4, 5
    # O . .   # Move 6, game over (player 0 wins).

    for i in range(1, 7):
      self.assertEqual(time_step.step_type.mid(), True)
      time_step = env.step([i])
    self.assertEqual(time_step.step_type.last(), True)

  def test_spec_fields(self):
    env = rl_environment.Environment("tic_tac_toe")
    env_spec = env.observation_spec()
    action_spec = env.action_spec()

    ttt_max_actions = 9
    ttt_normalized_info_set_shape = (27,)

    self.assertEqual(action_spec["num_actions"], ttt_max_actions)
    self.assertEqual(env_spec["info_state"], ttt_normalized_info_set_shape)
    self.assertCountEqual(
        env_spec.keys(),
        ["current_player", "info_state", "serialized_state", "legal_actions"])
    self.assertCountEqual(action_spec.keys(),
                          ["dtype", "max", "min", "num_actions"])

  def test_full_game_simultaneous_move(self):
    env = rl_environment.Environment("goofspiel")
    _ = env.reset()
    time_step = env.step([0, 0])
    self.assertEqual(time_step.observations["current_player"],
                     rl_environment.SIMULTANEOUS_PLAYER_ID)
    self.assertLen(time_step.observations["info_state"], 2)
    self.assertLen(time_step.observations["legal_actions"], 2)
    self.assertLen(time_step.rewards, 2)
    self.assertLen(time_step.discounts, 2)
    self.assertLen(time_step.observations, 4)

    actions = [act[0] for act in time_step.observations["legal_actions"]]
    time_step = env.step(actions)
    self.assertEqual(time_step.step_type.mid(), True)

    while not time_step.last():
      actions = [act[0] for act in time_step.observations["legal_actions"]]
      time_step = env.step(actions)

  def test_set_and_get_state(self):
    env_ttt1 = rl_environment.Environment("tic_tac_toe")
    env_ttt2 = rl_environment.Environment("tic_tac_toe")
    env_kuhn1 = rl_environment.Environment("kuhn_poker", players=2)
    env_kuhn2 = rl_environment.Environment("kuhn_poker", players=3)

    env_ttt1.reset()
    env_ttt2.reset()
    env_kuhn1.reset()
    env_kuhn2.reset()

    # Transfering states between identical games should work.
    env_ttt1.set_state(env_ttt2.get_state)
    env_ttt2.set_state(env_ttt1.get_state)

    # Transfering states between different games or games with different
    # parameters should fail.
    with self.assertRaises(AssertionError):
      self.fail(env_ttt1.set_state(env_kuhn1.get_state))
    with self.assertRaises(AssertionError):
      self.fail(env_kuhn1.set_state(env_ttt1.get_state))

    with self.assertRaises(AssertionError):
      self.fail(env_kuhn1.set_state(env_kuhn2.get_state))
    with self.assertRaises(AssertionError):
      self.fail(env_kuhn2.set_state(env_kuhn1.get_state))


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Examples of negotiation scenarios -- useful for generating more examples.
"""

from open_spiel.python.games.chat_games.envs.utils import text

# negotiating rent (money)
SCENARIO_A_LIST = [
    'Hi {receiver},', 'I hope you are well,', 'I understand you have been a ' +
    'long time tenant with me, so I hate to increase rent, but as you know ' +
    'inflation has increased by 6 percent recently. In order to stay ' +
    'solvent I will need to increase your rent by 6 percent as well. I hope ' +
    'you understand my thinking.\n\nHow do you feel about this? Would you ' +
    'like to continue renting from me?', 'Best,', '{sender}']
SCENARIO_A = '\n\n'.join(text.wrap(SCENARIO_A_LIST))

# negotiating deadline extension (time)
SCENARIO_B_LIST = [
    'Dear {receiver},', 'I understand that my payment is due at the end of ' +
    'this month, but I will find it hard to come up with the money. Would it ' +
    'be possible to extend the due date by 1 week? This would allow me to ' +
    'come up with the necessary funds. As a concession, I would be willing to' +
    ' pay early next month.', 'How do you feel about this? Do you have any ' +
    'other alternatives that you would be happy with?', 'Best,', '{sender}']
SCENARIO_B = '\n\n'.join(text.wrap(SCENARIO_B_LIST))

# negotiating a trade (things)
SCENARIO_C_LIST = [
    'Hey {receiver},', 'Thanks for your interest in my baseball card  ' +
    'collection. I see you like my signed Babe Ruth special edition card. To ' +
    'be honest, I really like your signed Nolan Ryan jersey. I also like ' +
    'your signed Roger Clemens ball. Would you be interested in a trade? I ' +
    'have a few other things you might like to sweeten the deal: Ken Griffey '+
    'Jr baseball bat, Mike Trout signed card, ...', 'What do you think?',
    'Best,', '{sender}']
SCENARIO_C = '\n\n'.join(text.wrap(SCENARIO_C_LIST))



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for gin rummy."""


from absl.testing import absltest

import pyspiel
gin_rummy = pyspiel.gin_rummy


class GamesGinRummyTest(absltest.TestCase):

  def test_bindings(self):
    # gin_rummy submodule attributes
    self.assertEqual(gin_rummy.DEFAULT_NUM_RANKS, 13)
    self.assertEqual(gin_rummy.DEFAULT_NUM_SUITS, 4)
    self.assertEqual(gin_rummy.DEFAULT_NUM_CARDS, 52)
    self.assertEqual(gin_rummy.NUM_PLAYERS, 2)
    self.assertEqual(gin_rummy.MAX_POSSIBLE_DEADWOOD, 98)
    self.assertEqual(gin_rummy.MAX_NUM_DRAW_UPCARD_ACTIONS, 50)
    self.assertEqual(gin_rummy.DEFAULT_HAND_SIZE, 10)
    self.assertEqual(gin_rummy.WALL_STOCK_SIZE, 2)
    self.assertEqual(gin_rummy.DEFAULT_KNOCK_CARD, 10)
    self.assertEqual(gin_rummy.DEFAULT_GIN_BONUS, 25)
    self.assertEqual(gin_rummy.DEFAULT_UNDERCUT_BONUS, 25)
    self.assertEqual(gin_rummy.DRAW_UPCARD_ACTION, 52)
    self.assertEqual(gin_rummy.DRAW_STOCK_ACTION, 53)
    self.assertEqual(gin_rummy.PASS_ACTION, 54)
    self.assertEqual(gin_rummy.KNOCK_ACTION, 55)
    self.assertEqual(gin_rummy.MELD_ACTION_BASE, 56)
    self.assertEqual(gin_rummy.NUM_MELD_ACTIONS, 185)
    self.assertEqual(gin_rummy.NUM_DISTINCT_ACTIONS, 241)
    self.assertEqual(gin_rummy.OBSERVATION_TENSOR_SIZE, 644)
    # Game bindings
    game = pyspiel.load_game('gin_rummy')
    self.assertFalse(game.oklahoma())
    self.assertEqual(game.knock_card(), 10)
    # State bindings
    state = game.new_initial_state()
    self.assertEqual(state.current_phase(), gin_rummy.Phase.DEAL)
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
    self.assertIsNone(state.upcard())
    self.assertEqual(state.stock_size(), 52)
    self.assertEqual(state.hands(), [[], []])
    self.assertEqual(state.discard_pile(), [])
    self.assertEqual(state.deadwood(), [0, 0])
    self.assertEqual(state.knocked(), [False, False])
    self.assertEqual(state.pass_on_first_upcard(), [False, False])
    self.assertEqual(state.layed_melds(), [[], []])
    self.assertEqual(state.layoffs(), [])
    self.assertFalse(state.finished_layoffs())
    for _ in range(25):  # 25 actions is sufficient to deal out all cards
      state.apply_action(state.legal_actions()[0])
    p0_resampled_state = state.resample_from_infostate(
        0, pyspiel.UniformProbabilitySampler(0., 1.))
    self.assertEqual(state.observation_string(0),
                     p0_resampled_state.observation_string(0))
    p1_resampled_state = state.resample_from_infostate(
        1, pyspiel.UniformProbabilitySampler(0., 1.))
    self.assertEqual(state.observation_string(1),
                     p1_resampled_state.observation_string(1))

    # Utils
    utils = gin_rummy.GinRummyUtils(gin_rummy.DEFAULT_NUM_RANKS,
                                    gin_rummy.DEFAULT_NUM_SUITS,
                                    gin_rummy.DEFAULT_HAND_SIZE)
    self.assertEqual(utils.card_string(0), 'As')
    self.assertEqual(utils.hand_to_string([0, 1, 2]),
                     '+--------------------------+\n'
                     '|As2s3s                    |\n'
                     '|                          |\n'
                     '|                          |\n'
                     '|                          |\n'
                     '+--------------------------+\n')
    self.assertEqual(utils.card_int('As'), 0)
    self.assertEqual(utils.card_ints_to_card_strings([0, 1, 2]),
                     ['As', '2s', '3s'])
    self.assertEqual(utils.card_strings_to_card_ints(['As', '2s', '3s']),
                     [0, 1, 2])
    self.assertEqual(utils.card_value(0), 1)
    self.assertEqual(utils.total_card_value([50, 51]), 20)
    self.assertEqual(utils.total_card_value([[0, 1], [50, 51]]), 23)
    self.assertEqual(utils.card_rank(51), 12)
    self.assertEqual(utils.card_suit(51), 3)
    self.assertTrue(utils.is_consecutive([0, 1, 2]))
    self.assertTrue(utils.is_rank_meld([0, 13, 26]))
    self.assertTrue(utils.is_suit_meld([0, 1, 2]))
    self.assertEqual(utils.rank_melds([0, 1, 13, 26]), [[0, 13, 26]])
    self.assertEqual(utils.suit_melds([0, 5, 6, 7]), [[5, 6, 7]])
    self.assertEqual(utils.all_melds([0, 5, 6, 7, 13, 26]),
                     [[0, 13, 26], [5, 6, 7]])
    self.assertEqual(utils.all_meld_groups([0, 5, 6, 7, 13, 26]),
                     [[[0, 13, 26], [5, 6, 7]], [[5, 6, 7], [0, 13, 26]]])
    self.assertEqual(utils.best_meld_group([0, 5, 6, 7, 13, 26]),
                     [[0, 13, 26], [5, 6, 7]])
    self.assertEqual(utils.min_deadwood([0, 1, 2], 3), 0)
    self.assertEqual(utils.min_deadwood([0, 1, 2]), 0)
    self.assertEqual(utils.rank_meld_layoff([0, 13, 26]), 39)
    self.assertEqual(utils.suit_meld_layoffs([0, 1, 2]), [3])
    self.assertEqual(utils.legal_melds([0, 1, 2, 3], 10), [65, 66, 109])
    self.assertEqual(utils.legal_discards([0, 1, 2], 10), [0, 1, 2])
    self.assertEqual(utils.all_layoffs([65], [3]), [4])
    self.assertEqual(utils.meld_to_int([0, 1, 2]), 65)
    self.assertEqual(utils.int_to_meld[65], [0, 1, 2])


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run deep online mirror descent algorithm with Munchausen DQN agents."""

import os
from typing import Sequence

from absl import flags
import jax

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.mfg import utils
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import munchausen_deep_mirror_descent
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import factory
from open_spiel.python.utils import app
from open_spiel.python.utils import metrics

FLAGS = flags.FLAGS

flags.DEFINE_string("game_name", "mfg_crowd_modelling_2d", "Name of the game.")
_ENV_SETTING = flags.DEFINE_string(
    "env_setting",
    "crowd_modelling_2d_four_rooms",
    "Name of the game settings. If None, the game name will be used.",
)

# Training options.
_BATCH_SIZE = flags.DEFINE_integer(
    "batch_size", 128, "Number of transitions to sample at each learning step."
)
_LEARN_EVERY = flags.DEFINE_integer(
    "learn_every", 64, "Number of steps between learning updates."
)
_NUM_EPISODES_PER_ITERATION = flags.DEFINE_integer(
    "num_episodes_per_iteration",
    1000,
    "Number of training eepisodes for each iteration.",
)
flags.DEFINE_integer("num_iterations", 100, "Number of iterations.")
_EPSILON_DECAY_DURATION = flags.DEFINE_integer(
    "epsilon_decay_duration",
    100000,
    "Number of game steps over which epsilon is decayed.",
)
flags.DEFINE_float("epsilon_power", 1, "Power for the epsilon decay.")
flags.DEFINE_float("epsilon_start", 0.1, "Starting exploration parameter.")
flags.DEFINE_float("epsilon_end", 0.1, "Final exploration parameter.")
_DISCOUNT_FACTOR = flags.DEFINE_float(
    "discount_factor", 1.0, "Discount factor for future rewards."
)
_RESET_REPLAY_BUFFER_ON_UPDATE = flags.DEFINE_bool(
    "reset_replay_buffer_on_update",
    False,
    "Reset the replay buffer when the softmax policy is updated.",
)
flags.DEFINE_integer("seed", 42, "Training seed.")
# Evaluation options.
_EVAL_EVERY = flags.DEFINE_integer(
    "eval_every", 200, "Episode frequency at which the agents are evaluated."
)
# Network options.
_HIDDEN_LAYERS_SIZES = flags.DEFINE_list(
    "hidden_layers_sizes",
    ["128", "128"],
    "Number of hidden units in the avg-net and Q-net.",
)
_UPDATE_TARGET_NETWORK_EVERY = flags.DEFINE_integer(
    "update_target_network_every",
    200,
    "Number of steps between DQN target network updates.",
)
# Replay buffer options.
_REPLAY_BUFFER_CAPACITY = flags.DEFINE_integer(
    "replay_buffer_capacity", 40000, "Size of the replay buffer."
)
_MIN_BUFFER_SIZE_TO_LEARN = flags.DEFINE_integer(
    "min_buffer_size_to_learn",
    1000,
    "Number of samples in buffer before learning begins.",
)
# Loss and optimizer options.
flags.DEFINE_enum("optimizer", "adam", ["sgd", "adam"], "Optimizer.")
flags.DEFINE_float("learning_rate", 0.01, "Learning rate for inner rl agent.")
flags.DEFINE_enum("loss", "mse", ["mse", "huber"], "Loss function.")
flags.DEFINE_float("huber_loss_parameter", 1.0, "Parameter for Huber loss.")
flags.DEFINE_float("gradient_clipping", None, "Value to clip the gradient to.")
# Munchausen options.
flags.DEFINE_float("tau", 10, "Temperature parameter in Munchausen target.")
flags.DEFINE_float("alpha", 0.99, "Alpha parameter in Munchausen target.")
_WITH_MUNCHAUSEN = flags.DEFINE_bool(
    "with_munchausen", True, "If true, target uses Munchausen penalty terms."
)
# Logging options.
flags.DEFINE_bool("use_checkpoints", False, "Save/load neural network weights.")
_CHECKPOINT_DIR = flags.DEFINE_string(
    "checkpoint_dir", "/tmp/dqn_test", "Directory to save/load the agent."
)
_LOGDIR = flags.DEFINE_string(
    "logdir",
    None,
    "Logging dir to use for TF summary files. If None, the metrics will only "
    "be logged to stderr.",
)
_LOG_DISTRIBUTION = flags.DEFINE_bool(
    "log_distribution", False, "Enables logging of the distribution."
)


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")

  game = factory.create_game_with_setting(FLAGS.game_name, _ENV_SETTING.value)

  num_players = game.num_players()

  # Create the environments with uniform initial policy.
  uniform_policy = policy.UniformRandomPolicy(game)
  uniform_dist = distribution.DistributionPolicy(game, uniform_policy)

  envs = [
      rl_environment.Environment(  # pylint: disable=g-complex-comprehension
          game,
          mfg_distribution=uniform_dist,
          mfg_population=p,
          observation_type=rl_environment.ObservationType.OBSERVATION,
      )
      for p in range(num_players)
  ]

  env = envs[0]
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  # Create the agents.
  kwargs = {
      "alpha": FLAGS.alpha,
      "batch_size": _BATCH_SIZE.value,
      "discount_factor": _DISCOUNT_FACTOR.value,
      "epsilon_decay_duration": _EPSILON_DECAY_DURATION.value,
      "epsilon_end": FLAGS.epsilon_end,
      "epsilon_power": FLAGS.epsilon_power,
      "epsilon_start": FLAGS.epsilon_start,
      "gradient_clipping": FLAGS.gradient_clipping,
      "hidden_layers_sizes": [int(l) for l in _HIDDEN_LAYERS_SIZES.value],
      "huber_loss_parameter": FLAGS.huber_loss_parameter,
      "learn_every": _LEARN_EVERY.value,
      "learning_rate": FLAGS.learning_rate,
      "loss": FLAGS.loss,
      "min_buffer_size_to_learn": _MIN_BUFFER_SIZE_TO_LEARN.value,
      "optimizer": FLAGS.optimizer,
      "replay_buffer_capacity": _REPLAY_BUFFER_CAPACITY.value,
      "reset_replay_buffer_on_update": _RESET_REPLAY_BUFFER_ON_UPDATE.value,
      "seed": FLAGS.seed,
      "tau": FLAGS.tau,
      "update_target_network_every": _UPDATE_TARGET_NETWORK_EVERY.value,
      "with_munchausen": _WITH_MUNCHAUSEN.value,
  }
  agents = [
      munchausen_deep_mirror_descent.MunchausenDQN(
          p, info_state_size, num_actions, **kwargs
      )
      for p in range(num_players)
  ]

  # Metrics writer will also log the metrics to stderr.
  just_logging = _LOGDIR.value is None or jax.host_id() > 0
  writer = metrics.create_default_writer(
      logdir=_LOGDIR.value, just_logging=just_logging
  )

  # # Save the parameters.
  writer.write_hparams(kwargs)

  def logging_fn(it, episode, vals):
    writer.write_scalars(it * num_episodes_per_iteration + episode, vals)

  num_episodes_per_iteration = _NUM_EPISODES_PER_ITERATION.value
  md = munchausen_deep_mirror_descent.DeepOnlineMirrorDescent(
      game,
      envs,
      agents,
      eval_every=_EVAL_EVERY.value,
      num_episodes_per_iteration=num_episodes_per_iteration,
      logging_fn=logging_fn,
  )

  def log_metrics(it):
    """Logs the training metrics for each iteration."""
    initial_states = game.new_initial_states()
    pi_value = policy_value.PolicyValue(game, md.distribution, md.policy)
    m = {
        f"best_response/{state}": pi_value.eval_state(state)
        for state in initial_states
    }
    nash_conv_md = nash_conv.NashConv(game, md.policy).nash_conv()
    m["nash_conv_md"] = nash_conv_md
    if _LOG_DISTRIBUTION.value and _LOGDIR.value:
      # We log distribution directly to a Pickle file as it may be large for
      # logging as a metric.
      filename = os.path.join(_LOGDIR.value, f"distribution_{it}.pkl")
      utils.save_parametric_distribution(md.distribution, filename)
    logging_fn(it, 0, m)

  log_metrics(0)
  for it in range(1, FLAGS.num_iterations + 1):
    md.iteration()
    log_metrics(it)

  # Make sure all values were written.
  writer.flush()


if __name__ == "__main__":
  jax.config.parse_flags_with_absl()
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.discounted_cfr."""

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import discounted_cfr
from open_spiel.python.algorithms import expected_game_score
import pyspiel


class DiscountedCfrTest(absltest.TestCase):

  def test_discounted_cfr_on_kuhn(self):
    game = pyspiel.load_game("kuhn_poker")
    solver = discounted_cfr.DCFRSolver(game)
    for _ in range(300):
      solver.evaluate_and_update_policy()
    average_policy = solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * 2)
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3)

  def test_discounted_cfr_runs_against_leduc(self):
    game = pyspiel.load_game("leduc_poker")
    solver = discounted_cfr.DCFRSolver(game)
    for _ in range(10):
      solver.evaluate_and_update_policy()
    solver.average_policy()


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fishburn's Maximal lotteries method.

Based on https://en.wikipedia.org/wiki/Maximal_lotteries.
"""

from typing import List
import numpy as np
from open_spiel.python.algorithms import lp_solver
import pyspiel
from open_spiel.python.voting import base


class MaximalLotteriesVoting(base.AbstractVotingMethod):
  """Implements Copeland's method."""

  def __init__(self,
               iterative: bool = True,
               verbose: bool = False,
               zero_tolerance: float = 1e-6):
    self._iterative = iterative
    self._verbose = verbose
    self._zero_tolerance = zero_tolerance

  def name(self) -> str:
    return f"maximal_lotteries(iterative={self._iterative})"

  def _create_matrix_game(self, matrix: np.ndarray):
    return pyspiel.create_tensor_game([matrix, -matrix]).as_matrix_game()

  def _solve_game(
      self, margin_matrix: np.ndarray
    ) -> np.ndarray:
    matrix_game = self._create_matrix_game(margin_matrix)
    p0_sol, _, _, _ = lp_solver.solve_zero_sum_matrix_game(matrix_game)
    return p0_sol

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    margin_matrix = profile.margin_matrix()
    alternatives = profile.alternatives
    m = profile.num_alternatives()
    if self._verbose:
      print(f"Margin matrix: \n{margin_matrix}")
      print(f"Alternatives: {alternatives}")
    p0_sol = self._solve_game(margin_matrix)

    # For now define scores as the probabilities.
    scores = {}
    if not self._iterative:
      # and negligible noise to break ties
      noise = 1e-10 * np.random.uniform(size=m)
      for i in range(m):
        scores[alternatives[i]] = p0_sol[i] + noise[i]
      sorted_scores = sorted(scores.items(), key=lambda item: item[1])
      sorted_scores.reverse()
      outcome = base.RankOutcome()
      outcome.unpack_from(sorted_scores)
      return outcome
    else:
      # Continue to iteratively solve all the remaining subgames.
      return self._iterate(alternatives, margin_matrix, p0_sol)

  def _iterate(
      self,
      alternatives: List[base.AlternativeId],
      margin_matrix: np.ndarray,
      p0_sol: np.ndarray,
  ):
    remaining_alternatives = alternatives[:]
    leveled_ranking = []
    leveled_scores = []
    while remaining_alternatives:
      # Pull out the nonzero entries and make them winners of this level.
      m = len(remaining_alternatives)
      if self._verbose:
        print(f"\nRemaining alternatives: {remaining_alternatives}")
        cur_level = len(leveled_ranking)
        print(f"IML Level {cur_level}")
        print(f"Remaining alternatives: {remaining_alternatives}")
        print(f"Margin matrix: \n{margin_matrix}\n")
      if m == 1:
        leveled_ranking.append(remaining_alternatives[:])
        leveled_scores.append([1])
        break
      noise = 1e-10 * np.random.uniform(size=m)
      for i in range(m):
        p0_sol[i] += noise[i]
      values = -1 * np.ones(m, dtype=np.float64)
      level_winners_idxs = []
      for i in range(m):
        if p0_sol[i] > self._zero_tolerance:
          # print(f"p0_sol[{i}] = {p0_sol[i]}")
          level_winners_idxs.append(i)
          values[i] = p0_sol[i]
      num_level_winners = len(level_winners_idxs)
      assert num_level_winners >= 1
      indices = np.argsort(-values)
      level_winners_ranked = []
      level_winners_scores = []
      for j in range(num_level_winners):
        idx = int(indices[j])
        level_winners_ranked.append(remaining_alternatives[idx])
        level_winners_scores.append(p0_sol[idx])
      leveled_ranking.append(level_winners_ranked)
      leveled_scores.append(level_winners_scores)
      if self._verbose:
        print(f"Level winners: {level_winners_ranked}")
        print(f"Level scores: {level_winners_scores}")
      # Now, take them out of the margin matrix and remaining alternatives
      # Delete in reverse order.
      for j in range(num_level_winners):
        idx = level_winners_idxs[num_level_winners - 1 - j]
        del remaining_alternatives[idx]
        margin_matrix = np.delete(margin_matrix, (idx), axis=0)
        margin_matrix = np.delete(margin_matrix, (idx), axis=1)
      if len(remaining_alternatives) > 1:
        p0_sol = self._solve_game(margin_matrix)
    # Now bump up the scores by level, and put them in the outcome.
    scores = {}
    num_levels = len(leveled_ranking)
    if self._verbose:
      print(f"Num levels: {num_levels}")
    level_base_points = num_levels - 1
    for level in range(num_levels):
      for j in range(len(leveled_ranking[level])):
        alternative = leveled_ranking[level][j]
        score = level_base_points + leveled_scores[level][j]
        scores[alternative] = score
      level_base_points -= 1
    sorted_scores = sorted(scores.items(), key=lambda item: item[1])
    sorted_scores.reverse()
    outcome = base.RankOutcome()
    outcome.unpack_from(sorted_scores)
    return outcome



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Weighted Voting Games.

A weighted voting game is a game where every player i has a weight w_i, and 
there is a fixed quota q, the characteristic function for coalition c is:

    v(c) = 1 if sum_{i in c} w_i > q,
           0 otherwise.

For more detail, see Chapter 4 of "Computational Aspects of Cooperative
Game Theory" text book by Georgios Chalkiadakis, Edith Elkind, and Michael
Wooldridge.
"""

import numpy as np
from open_spiel.python.coalitional_games import coalitional_game


class WeightedVotingGame(coalitional_game.CoalitionalGame):
  """Weighted Voting Game."""

  def __init__(self, weights: np.ndarray, quota: float):
    super().__init__(num_players=len(weights))
    assert len(weights.shape) == 1
    self._weights = weights
    self._quota = quota

  def coalition_value(self, coalition: np.ndarray) -> float:
    assert len(coalition) == self._num_players
    total_weight = np.inner(coalition, self._weights)
    if total_weight > self._quota:
      return 1.0
    else:
      return 0.0



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A basic AlphaZero implementation.

This implements the AlphaZero training algorithm. It spawns N actors which feed
trajectories into a replay buffer which are consumed by a learner. The learner
generates new weights, saves a checkpoint, and tells the actors to update. There
are also M evaluators running games continuously against a standard MCTS+Solver,
though each at a different difficulty (ie number of simulations for MCTS).

Due to the multi-process nature of this algorithm the logs are written to files,
one per process. The learner logs are also output to stdout. The checkpoints are
also written to the same directory.

Links to relevant articles/papers:
  https://deepmind.com/blog/article/alphago-zero-starting-scratch has an open
    access link to the AlphaGo Zero nature paper.
  https://deepmind.com/blog/article/alphazero-shedding-new-light-grand-games-chess-shogi-and-go
    has an open access link to the AlphaZero science paper.
"""

import collections
import datetime
import functools
import itertools
import json
import os
import random
import sys
import tempfile
import time
import traceback

import numpy as np

from open_spiel.python.algorithms import mcts
from open_spiel.python.algorithms.alpha_zero import evaluator as evaluator_lib
from open_spiel.python.algorithms.alpha_zero import model as model_lib
import pyspiel
from open_spiel.python.utils import data_logger
from open_spiel.python.utils import file_logger
from open_spiel.python.utils import spawn
from open_spiel.python.utils import stats

# Time to wait for processes to join.
JOIN_WAIT_DELAY = 0.001


class TrajectoryState(object):
  """A particular point along a trajectory."""

  def __init__(self, observation, current_player, legals_mask, action, policy,
               value):
    self.observation = observation
    self.current_player = current_player
    self.legals_mask = legals_mask
    self.action = action
    self.policy = policy
    self.value = value


class Trajectory(object):
  """A sequence of observations, actions and policies, and the outcomes."""

  def __init__(self):
    self.states = []
    self.returns = None

  def add(self, information_state, action, policy):
    self.states.append((information_state, action, policy))


class Buffer(object):
  """A fixed size buffer that keeps the newest values."""

  def __init__(self, max_size):
    self.max_size = max_size
    self.data = []
    self.total_seen = 0  # The number of items that have passed through.

  def __len__(self):
    return len(self.data)

  def __bool__(self):
    return bool(self.data)

  def append(self, val):
    return self.extend([val])

  def extend(self, batch):
    batch = list(batch)
    self.total_seen += len(batch)
    self.data.extend(batch)
    self.data[:-self.max_size] = []

  def sample(self, count):
    return random.sample(self.data, count)


class Config(collections.namedtuple(
    "Config", [
        "game",
        "path",
        "learning_rate",
        "weight_decay",
        "train_batch_size",
        "replay_buffer_size",
        "replay_buffer_reuse",
        "max_steps",
        "checkpoint_freq",
        "actors",
        "evaluators",
        "evaluation_window",
        "eval_levels",

        "uct_c",
        "max_simulations",
        "policy_alpha",
        "policy_epsilon",
        "temperature",
        "temperature_drop",

        "nn_model",
        "nn_width",
        "nn_depth",
        "observation_shape",
        "output_size",

        "quiet",
    ])):
  """A config for the model/experiment."""
  pass


def _init_model_from_config(config):
  return model_lib.Model.build_model(
      config.nn_model,
      config.observation_shape,
      config.output_size,
      config.nn_width,
      config.nn_depth,
      config.weight_decay,
      config.learning_rate,
      config.path)


def watcher(fn):
  """A decorator to fn/processes that gives a logger and logs exceptions."""
  @functools.wraps(fn)
  def _watcher(*, config, num=None, **kwargs):
    """Wrap the decorated function."""
    name = fn.__name__
    if num is not None:
      name += "-" + str(num)
    with file_logger.FileLogger(config.path, name, config.quiet) as logger:
      print("{} started".format(name))
      logger.print("{} started".format(name))
      try:
        return fn(config=config, logger=logger, **kwargs)
      except Exception as e:
        logger.print("\n".join([
            "",
            " Exception caught ".center(60, "="),
            traceback.format_exc(),
            "=" * 60,
        ]))
        print("Exception caught in {}: {}".format(name, e))
        raise
      finally:
        logger.print("{} exiting".format(name))
        print("{} exiting".format(name))
  return _watcher


def _init_bot(config, game, evaluator_, evaluation):
  """Initializes a bot."""
  noise = None if evaluation else (config.policy_epsilon, config.policy_alpha)
  return mcts.MCTSBot(
      game,
      config.uct_c,
      config.max_simulations,
      evaluator_,
      solve=False,
      dirichlet_noise=noise,
      child_selection_fn=mcts.SearchNode.puct_value,
      verbose=False,
      dont_return_chance_node=True)


def _play_game(logger, game_num, game, bots, temperature, temperature_drop):
  """Play one game, return the trajectory."""
  trajectory = Trajectory()
  actions = []
  state = game.new_initial_state()
  random_state = np.random.RandomState()
  logger.opt_print(" Starting game {} ".format(game_num).center(60, "-"))
  logger.opt_print("Initial state:\n{}".format(state))
  while not state.is_terminal():
    if state.is_chance_node():
      # For chance nodes, rollout according to chance node's probability
      # distribution
      outcomes = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes)
      action = random_state.choice(action_list, p=prob_list)
      action_str = state.action_to_string(state.current_player(), action)
      actions.append(action_str)
      state.apply_action(action)
    else:
      root = bots[state.current_player()].mcts_search(state)
      policy = np.zeros(game.num_distinct_actions())
      for c in root.children:
        policy[c.action] = c.explore_count
      policy = policy**(1 / temperature)
      policy /= policy.sum()
      if len(actions) >= temperature_drop:
        action = root.best_child().action
      else:
        action = np.random.choice(len(policy), p=policy)
      trajectory.states.append(
          TrajectoryState(state.observation_tensor(), state.current_player(),
                          state.legal_actions_mask(), action, policy,
                          root.total_reward / root.explore_count))
      action_str = state.action_to_string(state.current_player(), action)
      actions.append(action_str)
      logger.opt_print("Player {} sampled action: {}".format(
          state.current_player(), action_str))
      state.apply_action(action)
  logger.opt_print("Next state:\n{}".format(state))

  trajectory.returns = state.returns()
  logger.print("Game {}: Returns: {}; Actions: {}".format(
      game_num, " ".join(map(str, trajectory.returns)), " ".join(actions)))
  return trajectory


def update_checkpoint(logger, queue, model, az_evaluator):
  """Read the queue for a checkpoint to load, or an exit signal."""
  path = None
  while True:  # Get the last message, ignore intermediate ones.
    try:
      path = queue.get_nowait()
    except spawn.Empty:
      break
  if path:
    logger.print("Inference cache:", az_evaluator.cache_info())
    logger.print("Loading checkpoint", path)
    model.load_checkpoint(path)
    az_evaluator.clear_cache()
  elif path is not None:  # Empty string means stop this process.
    return False
  return True


@watcher
def actor(*, config, game, logger, queue):
  """An actor process runner that generates games and returns trajectories."""
  logger.print("Initializing model")
  model = _init_model_from_config(config)
  logger.print("Initializing bots")
  az_evaluator = evaluator_lib.AlphaZeroEvaluator(game, model)
  bots = [
      _init_bot(config, game, az_evaluator, False),
      _init_bot(config, game, az_evaluator, False),
  ]
  for game_num in itertools.count():
    if not update_checkpoint(logger, queue, model, az_evaluator):
      return
    queue.put(_play_game(logger, game_num, game, bots, config.temperature,
                         config.temperature_drop))


@watcher
def evaluator(*, game, config, logger, queue):
  """A process that plays the latest checkpoint vs standard MCTS."""
  results = Buffer(config.evaluation_window)
  logger.print("Initializing model")
  model = _init_model_from_config(config)
  logger.print("Initializing bots")
  az_evaluator = evaluator_lib.AlphaZeroEvaluator(game, model)
  random_evaluator = mcts.RandomRolloutEvaluator()

  for game_num in itertools.count():
    if not update_checkpoint(logger, queue, model, az_evaluator):
      return

    az_player = game_num % 2
    difficulty = (game_num // 2) % config.eval_levels
    max_simulations = int(config.max_simulations * (10 ** (difficulty / 2)))
    bots = [
        _init_bot(config, game, az_evaluator, True),
        mcts.MCTSBot(
            game,
            config.uct_c,
            max_simulations,
            random_evaluator,
            solve=True,
            verbose=False,
            dont_return_chance_node=True)
    ]
    if az_player == 1:
      bots = list(reversed(bots))

    trajectory = _play_game(logger, game_num, game, bots, temperature=1,
                            temperature_drop=0)
    results.append(trajectory.returns[az_player])
    queue.put((difficulty, trajectory.returns[az_player]))

    logger.print("AZ: {}, MCTS: {}, AZ avg/{}: {:.3f}".format(
        trajectory.returns[az_player],
        trajectory.returns[1 - az_player],
        len(results), np.mean(results.data)))


@watcher
def learner(*, game, config, actors, evaluators, broadcast_fn, logger):
  """A learner that consumes the replay buffer and trains the network."""
  logger.also_to_stdout = True
  replay_buffer = Buffer(config.replay_buffer_size)
  learn_rate = config.replay_buffer_size // config.replay_buffer_reuse
  logger.print("Initializing model")
  model = _init_model_from_config(config)
  logger.print("Model type: %s(%s, %s)" % (config.nn_model, config.nn_width,
                                           config.nn_depth))
  logger.print("Model size:", model.num_trainable_variables, "variables")
  save_path = model.save_checkpoint(0)
  logger.print("Initial checkpoint:", save_path)
  broadcast_fn(save_path)

  data_log = data_logger.DataLoggerJsonLines(config.path, "learner", True)

  stage_count = 7
  value_accuracies = [stats.BasicStats() for _ in range(stage_count)]
  value_predictions = [stats.BasicStats() for _ in range(stage_count)]
  game_lengths = stats.BasicStats()
  game_lengths_hist = stats.HistogramNumbered(game.max_game_length() + 1)
  outcomes = stats.HistogramNamed(["Player1", "Player2", "Draw"])
  evals = [Buffer(config.evaluation_window) for _ in range(config.eval_levels)]
  total_trajectories = 0

  def trajectory_generator():
    """Merge all the actor queues into a single generator."""
    while True:
      found = 0
      for actor_process in actors:
        try:
          yield actor_process.queue.get_nowait()
        except spawn.Empty:
          pass
        else:
          found += 1
      if found == 0:
        time.sleep(0.01)  # 10ms

  def collect_trajectories():
    """Collects the trajectories from actors into the replay buffer."""
    num_trajectories = 0
    num_states = 0
    for trajectory in trajectory_generator():
      num_trajectories += 1
      num_states += len(trajectory.states)
      game_lengths.add(len(trajectory.states))
      game_lengths_hist.add(len(trajectory.states))

      p1_outcome = trajectory.returns[0]
      if p1_outcome > 0:
        outcomes.add(0)
      elif p1_outcome < 0:
        outcomes.add(1)
      else:
        outcomes.add(2)

      replay_buffer.extend(
          model_lib.TrainInput(
              s.observation, s.legals_mask, s.policy, p1_outcome)
          for s in trajectory.states)

      for stage in range(stage_count):
        # Scale for the length of the game
        index = (len(trajectory.states) - 1) * stage // (stage_count - 1)
        n = trajectory.states[index]
        accurate = (n.value >= 0) == (trajectory.returns[n.current_player] >= 0)
        value_accuracies[stage].add(1 if accurate else 0)
        value_predictions[stage].add(abs(n.value))

      if num_states >= learn_rate:
        break
    return num_trajectories, num_states

  def learn(step):
    """Sample from the replay buffer, update weights and save a checkpoint."""
    losses = []
    for _ in range(len(replay_buffer) // config.train_batch_size):
      data = replay_buffer.sample(config.train_batch_size)
      losses.append(model.update(data))

    # Always save a checkpoint, either for keeping or for loading the weights to
    # the actors. It only allows numbers, so use -1 as "latest".
    save_path = model.save_checkpoint(
        step if step % config.checkpoint_freq == 0 else -1)
    losses = sum(losses, model_lib.Losses(0, 0, 0)) / len(losses)
    logger.print(losses)
    logger.print("Checkpoint saved:", save_path)
    return save_path, losses

  last_time = time.time() - 60
  for step in itertools.count(1):
    for value_accuracy in value_accuracies:
      value_accuracy.reset()
    for value_prediction in value_predictions:
      value_prediction.reset()
    game_lengths.reset()
    game_lengths_hist.reset()
    outcomes.reset()

    num_trajectories, num_states = collect_trajectories()
    total_trajectories += num_trajectories
    now = time.time()
    seconds = now - last_time
    last_time = now

    logger.print("Step:", step)
    logger.print(
        ("Collected {:5} states from {:3} games, {:.1f} states/s. "
         "{:.1f} states/(s*actor), game length: {:.1f}").format(
             num_states, num_trajectories, num_states / seconds,
             num_states / (config.actors * seconds),
             num_states / num_trajectories))
    logger.print("Buffer size: {}. States seen: {}".format(
        len(replay_buffer), replay_buffer.total_seen))

    save_path, losses = learn(step)

    for eval_process in evaluators:
      while True:
        try:
          difficulty, outcome = eval_process.queue.get_nowait()
          evals[difficulty].append(outcome)
        except spawn.Empty:
          break

    batch_size_stats = stats.BasicStats()  # Only makes sense in C++.
    batch_size_stats.add(1)
    data_log.write({
        "step": step,
        "total_states": replay_buffer.total_seen,
        "states_per_s": num_states / seconds,
        "states_per_s_actor": num_states / (config.actors * seconds),
        "total_trajectories": total_trajectories,
        "trajectories_per_s": num_trajectories / seconds,
        "queue_size": 0,  # Only available in C++.
        "game_length": game_lengths.as_dict,
        "game_length_hist": game_lengths_hist.data,
        "outcomes": outcomes.data,
        "value_accuracy": [v.as_dict for v in value_accuracies],
        "value_prediction": [v.as_dict for v in value_predictions],
        "eval": {
            "count": evals[0].total_seen,
            "results": [sum(e.data) / len(e) if e else 0 for e in evals],
        },
        "batch_size": batch_size_stats.as_dict,
        "batch_size_hist": [0, 1],
        "loss": {
            "policy": float(losses.policy),
            "value": float(losses.value),
            "l2reg": float(losses.l2),
            "sum": float(losses.total),
        },
        "cache": {  # Null stats because it's hard to report between processes.
            "size": 0,
            "max_size": 0,
            "usage": 0,
            "requests": 0,
            "requests_per_s": 0,
            "hits": 0,
            "misses": 0,
            "misses_per_s": 0,
            "hit_rate": 0,
        },
    })
    logger.print()

    if config.max_steps > 0 and step >= config.max_steps:
      break

    broadcast_fn(save_path)


def alpha_zero(config: Config):
  """Start all the worker processes for a full alphazero setup."""
  game = pyspiel.load_game(config.game)
  config = config._replace(
      observation_shape=game.observation_tensor_shape(),
      output_size=game.num_distinct_actions())

  print("Starting game", config.game)
  if game.num_players() != 2:
    sys.exit("AlphaZero can only handle 2-player games.")
  game_type = game.get_type()
  if game_type.reward_model != pyspiel.GameType.RewardModel.TERMINAL:
    raise ValueError("Game must have terminal rewards.")
  if game_type.dynamics != pyspiel.GameType.Dynamics.SEQUENTIAL:
    raise ValueError("Game must have sequential turns.")

  path = config.path
  if not path:
    path = tempfile.mkdtemp(prefix="az-{}-{}-".format(
        datetime.datetime.now().strftime("%Y-%m-%d-%H-%M"), config.game))
    config = config._replace(path=path)

  if not os.path.exists(path):
    os.makedirs(path)
  if not os.path.isdir(path):
    sys.exit("{} isn't a directory".format(path))
  print("Writing logs and checkpoints to:", path)
  print("Model type: %s(%s, %s)" % (config.nn_model, config.nn_width,
                                    config.nn_depth))

  with open(os.path.join(config.path, "config.json"), "w") as fp:
    fp.write(json.dumps(config._asdict(), indent=2, sort_keys=True) + "\n")

  actors = [spawn.Process(actor, kwargs={"game": game, "config": config,
                                         "num": i})
            for i in range(config.actors)]
  evaluators = [spawn.Process(evaluator, kwargs={"game": game, "config": config,
                                                 "num": i})
                for i in range(config.evaluators)]

  def broadcast(msg):
    for proc in actors + evaluators:
      proc.queue.put(msg)

  try:
    learner(game=game, config=config, actors=actors,  # pylint: disable=missing-kwoa
            evaluators=evaluators, broadcast_fn=broadcast)
  except (KeyboardInterrupt, EOFError):
    print("Caught a KeyboardInterrupt, stopping early.")
  finally:
    broadcast("")
    # for actor processes to join we have to make sure that their q_in is empty,
    # including backed up items
    for proc in actors:
      while proc.exitcode is None:
        while not proc.queue.empty():
          proc.queue.get_nowait()
        proc.join(JOIN_WAIT_DELAY)
    for proc in evaluators:
      proc.join()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Polynomial Weights algorithm for MFGs."""

from typing import Optional
import numpy as np
from open_spiel.python.mfg.algorithms import utils
from open_spiel.python.mfg.algorithms.regret import regret_minimizer


def polynomial_weight_update(weights, rewards, eta):
  return weights * (1 + eta * rewards)


class PolynomialWeightAlgorithm(regret_minimizer.RegretMinimizer):
  """Implements the Polynomial Weight Algorithm Regret minimizer.

  This is an external-regret minimizer, adapted here to the Mean-Field,
  Partially-Observable case.

  References: Muller et al, https://arxiv.org/abs/2111.08350, and
    Blum et al, https://www.cs.cmu.edu/~avrim/ML10/regret-chapter.pdf
  """

  def __init__(
      self,
      game,
      policies,
      eta: Optional[float] = None,
      regret_steps_per_step: int = 1,
      rho_tol: float = 1e-4,
      compress_nus: bool = True,
      compress_lbd: float = 0.0,
      compress_every: int = 1,
      stop_early: bool = True,
      stop_regret_threshold: float = 1e-3,
      value_estimator=utils.sample_value,
      value_estimation_n: int = 1,
      compute_internal_regret: bool = False,
  ):
    super().__init__(
        game,
        policies,
        regret_steps_per_step=regret_steps_per_step,
        rho_tol=rho_tol,
        compress_nus=compress_nus,
        compress_every=compress_every,
        compress_lbd=compress_lbd,
        stop_early=stop_early,
        stop_regret_threshold=stop_regret_threshold,
        value_estimator=value_estimator,
        value_estimation_n=value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )
    if self._compute_internal_regret:
      self._ws = [np.ones(len(policies)) for _ in range(len(policies))]
      self._p = np.ones(len(policies)) / (1.0 * len(policies))
    else:
      self._w = np.ones(len(policies))

    if eta is None:
      assert regret_steps_per_step is not None, (
          "Both `eta` and "
          "`regret_steps_per_step` were "
          "None, whereas our algorithm "
          "requires either value to be "
          "set."
      )
      self.compute_optimal_eta()
    else:
      self._eta = eta

    self._compress_every = compress_every

  def get_all_w_nus(self):
    assert self._compute_internal_regret
    return [w / np.sum(w) for w in self._ws]

  def get_nu(self):
    if self._compute_internal_regret:
      return np.sum(
          self._p.reshape(-1, 1) * np.array(self.get_all_w_nus()), axis=0
      )
    else:
      return self._w / np.sum(self._w)

  def _update_weights(self, rewards):
    if self._compute_internal_regret:
      self._ws = [
          w * (1 + self._eta * rewards * p) for w, p in zip(self._ws, self._p)
      ]
      self.compute_p()
    else:
      self._w = self._w * (1 + self._eta * rewards)

  def step(self, welfare_bonus=0.0):
    rewards = np.zeros(len(self._policies))
    nu = self.get_nu()
    assert np.all(nu >= 0.0) and (np.abs(np.sum(nu) - 1) < 1e-8)
    self._nus.append(nu)
    self._nu_weights.append(1.0)

    mu = utils.MixedDistribution(self._policy_mus, nu)
    for _ in range(self._value_estimation_n):
      for index, policy in enumerate(self._policies):
        rewards[index] += self._value_estimator(policy, mu, self._game)
    rewards /= self._value_estimation_n

    self._update_weights(rewards)

    welfare = np.sum(np.array(rewards) * np.array(nu))

    self._rewards.append(rewards + welfare_bonus * welfare * nu)
    self._true_rewards.append(rewards)

  def compute_optimal_eta(self):
    if self._regret_steps_per_step is not None:
      self._eta = min(
          np.sqrt(np.log(len(self._policies)) / self._regret_steps_per_step),
          0.5,
      )

  def reset(self, policies):
    if self._compute_internal_regret:
      self._ws = [np.ones(len(policies)) for _ in range(len(policies))]
      self._p = np.ones(len(policies)) / (1.0 * len(policies))
    else:
      self._w = np.ones(len(policies))
    self._policies = policies
    self._nus = []
    self._rewards = []
    self._true_rewards = []
    self._policy_mus = []
    self._nu_weights = []
    self.update_policy_mus()
    self.compute_optimal_eta()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Regret Matching Approximate Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import numpy as np


class Solver(object):
  """Regret-matching Solver."""

  def __init__(self, optimism=True, discount=False, rnd_init=False, seed=None,
               **kwargs):
    """Ctor."""
    del kwargs
    self.num_players = None
    self.lrs = None
    self.optimism = optimism
    self.discount = discount
    self.rnd_init = rnd_init
    self.has_aux = True
    self.aux_errors = []

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if self.rnd_init:
      init_dist = self.random.rand(num_strats)
    else:
      init_dist = np.ones(num_strats)
    init_dist /= init_dist.sum()
    init_regret = np.zeros(num_strats)
    return (init_dist, init_regret)

  def record_aux_errors(self, grads):
    """Record errors for the auxiliary variables."""
    grad_regret = grads[1]
    self.aux_errors.append([np.linalg.norm(grad_regret)])

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients (and exploitabilities) for all parameters.

    Args:
      params: tuple of params (dist, regret), see regmatch.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      tuple of gradients (grad_dist, grad_regret), see ate.gradients
      unregularized exploitability (stochastic estimate)
      solver exploitability (stochastic estimate) - NaN
    """
    return gradients(*params, payoff_matrices)

  def exploitability(self, params, payoff_matrices):
    """Regret matching does not minimize any exploitability so return NaN.

    Args:
      params: tuple of params (dist,)
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      np.NaN
    """
    del params
    del payoff_matrices
    return np.nan

  def update(self, params, grads, t):
    """Update cumulative regret and strategy (dist).

    Args:
      params: tuple of variables to be updated (dist, regret)
      grads: tuple of variable gradients (grad_dist, grad_regret)
      t: int, solver iteration (not used)
    Returns:
      new_params: tuple of update params (new_dist, new_regret)
    """
    dist, regret = params
    regret_delta = grads[1]
    if self.discount:
      gamma = t / float(t + 1)
    else:
      gamma = 1
    new_regret = gamma * regret + regret_delta
    new_clipped_regrets = np.clip(new_regret + self.optimism * regret_delta,
                                  0.,
                                  np.inf)
    if np.sum(new_clipped_regrets) > 0:
      new_dist = new_clipped_regrets / new_clipped_regrets.sum()
    else:
      new_dist = np.ones_like(dist) / dist.size
    new_params = (new_dist, new_regret)
    return new_params


def gradients(dist, regret, payoff_matrices):
  """Computes regret delta to be added to regret in update.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    regret: 1-d np.array (same shape as dist), current estimate of regrets
    payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
  Returns:
    deltas w.r.t. (dist, regret) as tuple
    unregularized exploitability (stochastic estimate)
    solver exploitability (stochastic estimate) - NaN
  """
  del regret

  nabla = payoff_matrices[0].dot(dist)
  utility = nabla.dot(dist)

  grad_dist = np.nan * np.ones_like(dist)
  grad_regret = nabla - utility

  unreg_exp = np.max(nabla) - nabla.dot(dist)

  return (grad_dist, grad_regret), unreg_exp, np.nan



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Boltzmann Q learning agent.

This algorithm is a variation of Q learning that uses action selection
based on boltzmann probability interpretation of Q-values.

For more details, see equation (2) page 2 in
   https://arxiv.org/pdf/1109.1528.pdf
"""

import numpy as np

from open_spiel.python import rl_tools
from open_spiel.python.algorithms import tabular_qlearner


class BoltzmannQLearner(tabular_qlearner.QLearner):
  """Tabular Boltzmann Q-Learning agent.

  See open_spiel/python/examples/tic_tac_toe_qlearner.py for an usage example.

  The tic_tac_toe example uses the standard Qlearner. Using the
  BoltzmannQlearner is
  identical and only differs in the initialization of the agents.
  """

  def __init__(self,
               player_id,
               num_actions,
               step_size=0.1,
               discount_factor=1.0,
               temperature_schedule=rl_tools.ConstantSchedule(.5),
               centralized=False):
    super().__init__(
        player_id,
        num_actions,
        step_size=step_size,
        discount_factor=discount_factor,
        epsilon_schedule=temperature_schedule,
        centralized=centralized)

  def _softmax(self, info_state, legal_actions, temperature):
    """Action selection based on boltzmann probability interpretation of Q-values.

    For more details, see equation (2) page 2 in
    https://arxiv.org/pdf/1109.1528.pdf

    Args:
        info_state: hashable representation of the information state.
        legal_actions: list of actions at `info_state`.
        temperature: temperature used for softmax.

    Returns:
        A valid soft-max selected action and valid action probabilities.
    """
    probs = np.zeros(self._num_actions)

    if temperature > 0.0:
      probs += [
          np.exp((1 / temperature) * self._q_values[info_state][i])
          for i in range(self._num_actions)
      ]
      probs /= np.sum(probs)
    else:
      # Temperature = 0 causes normal greedy action selection
      greedy_q = max([self._q_values[info_state][a] for a in legal_actions])
      greedy_actions = [
          a for a in legal_actions if self._q_values[info_state][a] == greedy_q
      ]

      probs[greedy_actions] += 1 / len(greedy_actions)

    action = np.random.choice(range(self._num_actions), p=probs)
    return action, probs

  def _get_action_probs(self, info_state, legal_actions, epsilon):
    """Returns a selected action and the probabilities of legal actions."""
    return self._softmax(info_state, legal_actions, temperature=epsilon)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stochastic Gradient Descent (Adam) Approx. Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import jax
import jax.numpy as jnp

import numpy as np

import optax

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.nonsymmetric import exploitability as exp
from open_spiel.python.algorithms.adidas_utils.helpers.nonsymmetric import updates


class Solver(updates.Solver):
  """Adam Solver."""

  def __init__(self, temperature=0., proj_grad=True, euclidean=False,
               lrs=(1e-1,), rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    super().__init__(proj_grad, euclidean, rnd_init, seed)
    if temperature < 0.:
      raise ValueError('temperature must be non-negative')
    self.temperature = temperature
    self.lrs = lrs
    self.num_estimates = 2

    if temperature > 0:
      self.eps = np.exp(-1 / temperature)  # ensure dist[i] >= eps / dim(dist)
    else:
      self.eps = 0.
    self.update = lambda *args: self.descent_step(*args, eps=self.eps)

    self.opt = optax.adam(learning_rate=lrs[0])
    self.opt_state = self.opt.init(jnp.zeros(1))

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if len(num_strats) != num_players:
      raise ValueError('Must specify num strategies for each player')
    init_dist = []
    for num_strats_i in num_strats:
      if self.rnd_init:
        init_dist_i = self.random.rand(num_strats_i)
      else:
        init_dist_i = np.ones(num_strats_i)
      init_dist_i /= init_dist_i.sum()
      init_dist_i = simplex.project_to_interior(init_dist_i, self.eps)
      init_dist.append(init_dist_i)

    init_params = [
        jnp.array(dist_to_logits(init_dist_i)) for init_dist_i in init_dist
    ]

    self.opt_state = self.opt.init(init_params)

    return (init_dist,)

  def descent_step(self, params, grads, t, eps=0.):
    """Projected gradient descent on exploitability using Euclidean projection.

    Args:
      params: tuple of variables to be updated (dist,)
      grads: tuple of variable gradients (grad_dist,)
      t: int, solver iteration (unused)
      eps: float > 0, force all probabilities >= eps / dim(dist)
    Returns:
      new_params: tuple of update params (new_dist,)
    """
    del t
    del eps

    dist = params[0]
    grads_dist = grads[0]

    dist_jnp = [jnp.array(dist_i) for dist_i in dist]
    grads_dist_jnp = [jnp.array(grad_i) for grad_i in grads_dist]

    # map dist to logits and grads to grad_logits using jacobian
    logits = [dist_to_logits(dist_i) for dist_i in params[0]]
    grads_logits = [
        jax.jvp(dist_to_logits, [dist_i], [grads_i])[1]
        for dist_i, grads_i in zip(dist_jnp, grads_dist_jnp)
    ]

    opt_updates, self.opt_state = self.opt.update(grads_logits,
                                                  self.opt_state,
                                                  logits)

    new_logits = optax.apply_updates(logits, opt_updates)

    new_dist = [logits_to_dist(logits) for logits in new_logits]
    new_dist = [np.array(dist_i) for dist_i in new_dist]

    return (new_dist,)

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return exploitability.

    Args:
      params: tuple of params (dist,), see sgd.gradients
      payoff_matrices: 2 dictionaries with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    Returns:
      float, exploitability of current dist
      unregularized exploitability (stochastic estimate)
      shannon regularized exploitability (stochastic estimate)
    """
    return gradients(*params, payoff_matrices, self.num_players,
                     self.temperature, self.proj_grad)

  def exploitability(self, params, payoff_matrices):
    """Compute and return exploitability.

    Args:
      params: tuple of params (dist,), see sgd.gradients
      payoff_matrices: (>=2 x A x A) np.array, payoffs for each joint action
    Returns:
      float, exploitability as avg squared norm of projected-gradient
    """
    return exp.grad_norm_exploitability(params, payoff_matrices, eta=1.,
                                        temperature=self.temperature)


def logits_to_dist(logits):
  logits_ext = jnp.append(logits, 0.)
  payoff = jax.nn.softmax(logits_ext)
  return payoff


def dist_to_logits(dist):
  # dist[-1] = exp(logits[-1]) / Z = exp(0) / Z
  z = 1 / dist[-1]
  logits = jnp.log(dist[:-1] * z)
  return logits


def gradients(dist, payoff_matrices, num_players, temperature=0.,
              proj_grad=True):
  """Computes exploitablity gradient.

  Args:
    dist: list of 1-d np.arrays, current estimate of nash distribution
    payoff_matrices: 2 dictionaries with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    num_players: int, number of players, in case payoff_matrices is abbreviated
    temperature: non-negative float, default 0.
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist) as tuple
    unregularized exploitability (stochastic estimate)
    shannon regularized exploitability (stochastic estimate)
  """
  # first compute projected gradients (for every player, for each sample a & b)
  # if consulting paper https://arxiv.org/abs/2310.06689, code assumes eta_k = 1
  tau = temperature

  pgs = []
  for i in range(num_players):

    pg_i_a = np.zeros_like(dist[i])
    pg_i_b = np.zeros_like(dist[i])

    for j in range(num_players):
      if j == i:
        continue
      if i < j:
        hess_i_ij_a = payoff_matrices[0][(i, j)][0]
        hess_i_ij_b = payoff_matrices[1][(i, j)][0]
      else:
        hess_i_ij_a = payoff_matrices[0][(j, i)][1].T
        hess_i_ij_b = payoff_matrices[1][(j, i)][1].T

      pg_i_a_est = simplex.project_grad(hess_i_ij_a.dot(dist[j]))
      pg_i_b_est = simplex.project_grad(hess_i_ij_b.dot(dist[j]))

      pg_i_a += pg_i_a_est / float(num_players - 1)
      pg_i_b += pg_i_b_est / float(num_players - 1)

    pgs.append((pg_i_a, pg_i_b))

  # then construct unbiased stochastic gradient
  grad_dist = []
  unreg_exp = []
  reg_exp = []

  for i in range(num_players):

    grad_dist_i = np.zeros_like(dist[i])

    for j in range(num_players):
      pg_j_a = pgs[j][0]
      pg_j_b = pgs[j][1]
      if tau > 0.:
        log_dist_safe = np.clip(np.log(dist[j]), -40, 0)
        entr_grad_proj = simplex.project_grad(-tau * (log_dist_safe + 1))
      else:
        entr_grad_proj = 0.
      pg_j_a_entr = pg_j_a + entr_grad_proj
      pg_j_b_entr = pg_j_b + entr_grad_proj

      if j == i:
        if tau > 0.:
          hess_j_ij_a = -tau * np.diag(1. / dist[j])
        else:
          hess_j_ij_a = np.diag(np.zeros_like(dist[j]))
        unreg_exp_i = np.dot(pg_j_a, pg_j_b)
        reg_exp_i = np.dot(pg_j_a_entr, pg_j_b_entr)
        unreg_exp.append(unreg_exp_i)
        reg_exp.append(reg_exp_i)
      elif i < j:
        hess_j_ij_a = payoff_matrices[0][(i, j)][1]
      else:
        hess_j_ij_a = payoff_matrices[0][(j, i)][0].T

      grad_dist_i += 2. * hess_j_ij_a.dot(pg_j_b_entr)

    if proj_grad:
      grad_dist_i = simplex.project_grad(grad_dist_i)

    grad_dist.append(grad_dist_i)

  return (grad_dist,), np.mean(unreg_exp), np.mean(reg_exp)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Runs mean field proximal policy optimaztion agents."""

# pylint: disable=consider-using-from-import

import logging
import os
import time

from absl import flags
import numpy as np
import torch
import torch.optim as optim
from torch.utils.tensorboard import SummaryWriter

from open_spiel.python import policy as policy_std
from open_spiel.python import rl_environment
from open_spiel.python.mfg import utils
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import Agent as mfg_ppo_agent
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import calculate_advantage
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import calculate_explotability
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import learn
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import Policy as mfg_ppo_policy
from open_spiel.python.mfg.algorithms.pytorch.mfg_proximal_policy_optimization import rollout
from open_spiel.python.mfg.games import factory
from open_spiel.python.utils import app


FLAGS = flags.FLAGS


flags.DEFINE_integer("seed", default=0, help="Set a random seed.")
flags.DEFINE_string(
    "exp_name", default="mf-ppo", help="Set the name of this experiment"
)
flags.DEFINE_string(
    "game_setting",
    default="crowd_modelling_2d_four_rooms",
    help=(
        "Set the game to benchmark options:(crowd_modelling_2d_four_rooms)     "
        "                and (crowd_modelling_2d_maze)"
    ),
)
flags.DEFINE_float("lr", default=1e-3, help="Learning rate of the optimizer")
flags.DEFINE_integer(
    "num_episodes",
    default=5,
    help=(
        "set the number of episodes                      of to collect per"
        " rollout"
    ),
)
flags.DEFINE_integer(
    "update_episodes",
    default=20,
    help="set the number of episodes                      of the inner loop",
)
flags.DEFINE_integer(
    "update_iterations",
    default=100,
    help=(
        "Set the number of global                      update steps of the"
        " outer loop"
    ),
)
flags.DEFINE_string(
    "optimizer", default="Adam", help="Set the optimizer (Adam) or (SGD)"
)
flags.DEFINE_boolean(
    "cuda", default=False, help="Use Gpu to run the experiment"
)

# MFPPO parameters
flags.DEFINE_float("gamma", default=0.9, help="set discount factor gamma")
flags.DEFINE_integer(
    "num_minibatches", default=5, help="the number of mini-batches"
)
flags.DEFINE_integer(
    "update_epochs", default=5, help="the K epochs to update the policy"
)
flags.DEFINE_float(
    "clip_coef", default=0.2, help="the surrogate clipping coefficient"
)
flags.DEFINE_float("ent_coef", default=0.01, help="coefficient of the entropy")
flags.DEFINE_float(
    "max_grad_norm",
    default=5,
    help="the maximum norm for the gradient clipping",
)
flags.DEFINE_float(
    "alpha",
    default=0.5,
    help=(
        "Set alpha to controll the iteration                    and epsiode"
        " policy updates"
    ),
)
flags.DEFINE_float(
    "eps_eps", default=0.2, help="eps to update the episode learned policy"
)
flags.DEFINE_float(
    "itr_eps", default=0.05, help="eps to update the episode learned policy"
)


def set_seed(seed):
  """Set the random seed for reproducibility."""
  np.random.seed(seed)
  torch.manual_seed(seed)
  torch.cuda.manual_seed(seed)
  os.environ["PYTHONHASHSEED"] = str(seed)
  print(f"Random seed set as {seed}")


def main(unused_argv):
  """Main function to run the experiment."""

  # Set the random seed for reproducibility
  set_seed(FLAGS.seed)

  # Set the device (in our experiments CPU vs GPU does not improve time at all)
  # we recommend CPU
  device = torch.device(
      "cuda" if torch.cuda.is_available() and FLAGS.cuda else "cpu"
  )

  # Set the name of the experiment's folder
  fname = "./mfppo_experiments/"

  # Log the experiments
  run_name = (
      f"{FLAGS.exp_name}_{FLAGS.game_setting}_{FLAGS.optimizer}_num_update_epochs_"
      "     "
      f" {FLAGS.update_epochs}_num_episodes_per_rollout_{FLAGS.num_episodes}_number_of_mini_batches_"
      "     "
      f" {FLAGS.num_minibatches}_{time.asctime(time.localtime(time.time()))}"
  )
  log_name = os.path.join(fname, run_name)
  tb_writer = SummaryWriter(log_name)
  logging.basicConfig(
      filename=log_name + "_log.txt",
      filemode="a",
      level=logging.DEBUG,
      force=True,
  )

  # Console handler
  console = logging.StreamHandler()
  console.setLevel(logging.ERROR)
  logging.getLogger("").addHandler(console)

  logger = logging.getLogger()
  logger.debug("Initialization")

  tb_writer.add_text(
      "hyperparameters",
      "|param|value|\n|-|-|\n%s"
      % "\n".join([f"|{key}|{value}" for key, value in vars(FLAGS).items()]),
  )
  # Create the game instance
  game = factory.create_game_with_setting(
      "mfg_crowd_modelling_2d", FLAGS.game_setting
  )

  # Set the initial policy to uniform and generate the distribution
  uniform_policy = policy_std.UniformRandomPolicy(game)
  mfg_dist = distribution.DistributionPolicy(game, uniform_policy)
  env = rl_environment.Environment(
      game, mfg_distribution=mfg_dist, mfg_population=0
  )

  # Set the environment seed for reproduciblility
  env.seed(FLAGS.seed)

  # Creat the agent and population policies
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]
  agent = mfg_ppo_agent(info_state_size, num_actions).to(device)
  ppo_policy = mfg_ppo_policy(game, agent, None, device)
  pop_agent = mfg_ppo_agent(info_state_size, num_actions).to(device)

  if FLAGS.optimizer == "Adam":
    optimizer_actor = optim.Adam(
        agent.actor.parameters(), lr=FLAGS.lr, eps=1e-5
    )
    optimizer_critic = optim.Adam(
        agent.critic.parameters(), lr=FLAGS.lr, eps=1e-5
    )
  else:
    optimizer_actor = optim.SGD(
        agent.actor.parameters(), lr=FLAGS.lr, momentum=0.9
    )
    optimizer_critic = optim.SGD(
        agent.critic.parameters(), lr=FLAGS.lr, momentum=0.9
    )

  # Used to log data for debugging
  steps = FLAGS.num_episodes * env.max_game_length
  episode_entropy = []
  total_entropy = []
  nash_con_vect = []
  eps_reward = []
  total_reward = []

  for k in range(FLAGS.update_iterations):
    for _ in range(FLAGS.update_episodes):
      # collect rollout data
      history = rollout(
          env, pop_agent, agent, FLAGS.num_episodes, steps, device
      )
      # store rewards and entropy for debugging
      episode_entropy.append(history["entropies"].mean().item())
      eps_reward.append(history["rewards"].sum().item() / FLAGS.num_episodes)
      # Calculate the advantage function
      adv, returns = calculate_advantage(
          FLAGS.gamma,
          True,
          history["rewards"],
          history["values"],
          history["dones"],
          device,
      )
      history["advantages"] = adv
      history["returns"] = returns
      # Update the learned policy and report loss for debugging
      v_loss = learn(
          history,
          optimizer_actor,
          optimizer_critic,
          agent,
          num_minibatches=FLAGS.num_minibatches,
          update_epochs=FLAGS.update_epochs,
          itr_eps=FLAGS.itr_eps,
          eps_eps=FLAGS.eps_eps,
          alpha=FLAGS.alpha,
          ent_coef=FLAGS.ent_coef,
          max_grad_norm=FLAGS.max_grad_norm,
      )

    # Collect and print the metrics
    total_reward.append(np.mean(eps_reward))
    total_entropy.append(np.mean(episode_entropy))

    print("Value_loss", v_loss.item())
    print("iteration num:", k + 1)
    print("Mean reward", total_reward[-1])

    # Update the iteration policy with the new policy
    pop_agent.load_state_dict(agent.state_dict())

    # Update the distribution
    distrib = distribution.DistributionPolicy(game, ppo_policy)

    # calculate the exploitability
    m = calculate_explotability(game, distrib, ppo_policy)
    nashc = m["nash_conv_ppo"]
    nash_con_vect.append(nashc)

    # log the results to tensor board
    tb_writer.add_scalar("initial_state_value", m["ppo_br/initial"], k + 1)
    tb_writer.add_scalar("rewards", total_reward[-1], k + 1)
    tb_writer.add_scalar("entorpy", total_entropy[-1], k + 1)
    tb_writer.add_scalar("nash_conv_ppo", nashc, k + 1)
    logger.debug(
        "ppo_br: %s, and nash_conv: %s, reward: %s, entropy: %s",
        m["ppo_br/initial"],
        nashc,
        total_reward[-1],
        total_entropy[-1],
    )
    print(
        "ppo_br: %s, and nash_conv: %s, reward: %s, entropy: %s"
        % (m["ppo_br/initial"], nashc, total_reward[-1], total_entropy[-1])
    )

    # Update the environment distribution
    env.update_mfg_distribution(distrib)

  # if lower than upper_nash we save the weights and distribution
  upper_nash = 300
  if nash_con_vect[-1] < upper_nash:
    # Save the distribution and weights for further analysis
    filename = os.path.join(fname, f"distribution_{run_name}.pkl")
    utils.save_parametric_distribution(distrib, filename)
    torch.save(
        agent.actor.state_dict(),
        fname
        + f"alpha_{FLAGS.alpha},                itr_eps_{FLAGS.itr_eps},"
        f" eps_eps_{FLAGS.eps_eps}_agent_actor_weights.pth",
    )
    torch.save(
        agent.critic.state_dict(),
        fname
        + f"alpha_{FLAGS.alpha},                itr_eps_{FLAGS.itr_eps},"
        f" eps_eps_{FLAGS.eps_eps}_agent_critic_weights.pth",
    )


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for base environments."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.games.chat_games.envs.base_envs import email_plain
from open_spiel.python.games.chat_games.envs.base_envs import email_with_tone
from open_spiel.python.games.chat_games.envs.base_envs import email_with_tone_info
from open_spiel.python.games.chat_games.envs.base_envs import schedule_meeting_with_info
from open_spiel.python.games.chat_games.envs.base_envs import trade_fruit_with_info
from open_spiel.python.games.chat_games.envs.utils import header


class BaseEnvsTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(base_env=email_plain),
      dict(base_env=email_with_tone),
      dict(base_env=email_with_tone_info),
      dict(base_env=schedule_meeting_with_info),
      dict(base_env=trade_fruit_with_info),
  ])
  def test_give_me_a_name(self, base_env):
    self.assertTrue(header.plain_header_is_valid(base_env.HEADER))


if __name__ == '__main__':
  absltest.main()



"""Abseil compiler options.

This is the source of truth for Abseil compiler options.  To modify Abseil
compilation options:

  (1) Edit the appropriate list in this file based on the platform the flag is
      needed on.
  (2) Run `<path_to_absl>/copts/generate_copts.py`.

The generated copts are consumed by configure_copts.bzl and
AbseilConfigureCopts.cmake.
"""

ABSL_GCC_FLAGS = [
    "-Wall",
    "-Wextra",
    "-Wcast-qual",
    "-Wconversion-null",
    "-Wformat-security",
    "-Wmissing-declarations",
    "-Wnon-virtual-dtor",
    "-Woverlength-strings",
    "-Wpointer-arith",
    "-Wundef",
    "-Wunused-local-typedefs",
    "-Wunused-result",
    "-Wvarargs",
    "-Wvla",  # variable-length array
    "-Wwrite-strings",
    # Don't define min and max macros (Build on Windows using gcc)
    "-DNOMINMAX",
]

ABSL_GCC_TEST_ADDITIONAL_FLAGS = [
    "-Wno-deprecated-declarations",
    "-Wno-missing-declarations",
    "-Wno-self-move",
    "-Wno-sign-compare",
    "-Wno-unused-function",
    "-Wno-unused-parameter",
    "-Wno-unused-private-field",
]

ABSL_LLVM_FLAGS = [
    "-Wall",
    "-Wextra",
    "-Wc++98-compat-extra-semi",
    "-Wcast-qual",
    "-Wconversion",
    "-Wdeprecated-pragma",
    "-Wfloat-overflow-conversion",
    "-Wfloat-zero-conversion",
    "-Wfor-loop-analysis",
    "-Wformat-security",
    "-Wgnu-redeclared-enum",
    "-Winfinite-recursion",
    "-Winvalid-constexpr",
    "-Wliteral-conversion",
    "-Wmissing-declarations",
    "-Woverlength-strings",
    "-Wpointer-arith",
    "-Wself-assign",
    "-Wshadow-all",
    "-Wshorten-64-to-32",
    "-Wsign-conversion",
    "-Wstring-conversion",
    "-Wtautological-overlap-compare",
    "-Wtautological-unsigned-zero-compare",
    "-Wundef",
    "-Wuninitialized",
    "-Wunreachable-code",
    "-Wunused-comparison",
    "-Wunused-local-typedefs",
    "-Wunused-result",
    "-Wvla",
    "-Wwrite-strings",
    # Warnings that are enabled by group warning flags like -Wall that we
    # explicitly disable.
    "-Wno-float-conversion",
    "-Wno-implicit-float-conversion",
    "-Wno-implicit-int-float-conversion",
    # Disable warnings on unknown warning flags (when warning flags are
    # unknown on older compiler versions)
    "-Wno-unknown-warning-option",
    # Don't define min and max macros (Build on Windows using clang)
    "-DNOMINMAX",
]

ABSL_LLVM_TEST_ADDITIONAL_FLAGS = [
    "-Wno-deprecated-declarations",
    "-Wno-implicit-int-conversion",
    "-Wno-missing-prototypes",
    "-Wno-missing-variable-declarations",
    "-Wno-shadow",
    "-Wno-shorten-64-to-32",
    "-Wno-sign-compare",
    "-Wno-sign-conversion",
    "-Wno-unreachable-code-loop-increment",
    "-Wno-unused-function",
    "-Wno-unused-member-function",
    "-Wno-unused-parameter",
    "-Wno-unused-private-field",
    "-Wno-unused-template",
    "-Wno-used-but-marked-unused",
    # gtest depends on this GNU extension being offered.
    "-Wno-gnu-zero-variadic-macro-arguments",
]

# /Wall with msvc includes unhelpful warnings such as C4711, C4710, ...
MSVC_BIG_WARNING_FLAGS = [
    "/W3",
]

MSVC_WARNING_FLAGS = [
    # Increase the number of sections available in object files
    "/bigobj",
    "/wd4005",  # macro-redefinition
    "/wd4068",  # unknown pragma
    # qualifier applied to function type has no meaning; ignored
    "/wd4180",
    # conversion from 'type1' to 'type2', possible loss of data
    "/wd4244",
    # conversion from 'size_t' to 'type', possible loss of data
    "/wd4267",
    # The decorated name was longer than the compiler limit
    "/wd4503",
    # forcing value to bool 'true' or 'false' (performance warning)
    "/wd4800",
]

MSVC_DEFINES = [
    "/DNOMINMAX",  # Don't define min and max macros (windows.h)
    # Don't bloat namespace with incompatible winsock versions.
    "/DWIN32_LEAN_AND_MEAN",
    # Don't warn about usage of insecure C functions.
    "/D_CRT_SECURE_NO_WARNINGS",
    "/D_SCL_SECURE_NO_WARNINGS",
    # Introduced in VS 2017 15.8, allow overaligned types in aligned_storage
    "/D_ENABLE_EXTENDED_ALIGNED_STORAGE",
]


def GccStyleFilterAndCombine(default_flags, test_flags):
  """Merges default_flags and test_flags for GCC and LLVM.

  Args:
    default_flags: A list of default compiler flags
    test_flags: A list of flags that are only used in tests

  Returns:
    A combined list of default_flags and test_flags, but with all flags of the
    form '-Wwarning' removed if test_flags contains a flag of the form
    '-Wno-warning'
  """
  remove = set(["-W" + f[5:] for f in test_flags if f[:5] == "-Wno-"])
  return [f for f in default_flags if f not in remove] + test_flags

COPT_VARS = {
    "ABSL_GCC_FLAGS": ABSL_GCC_FLAGS,
    "ABSL_GCC_TEST_FLAGS": GccStyleFilterAndCombine(
        ABSL_GCC_FLAGS, ABSL_GCC_TEST_ADDITIONAL_FLAGS),
    "ABSL_LLVM_FLAGS": ABSL_LLVM_FLAGS,
    "ABSL_LLVM_TEST_FLAGS": GccStyleFilterAndCombine(
        ABSL_LLVM_FLAGS, ABSL_LLVM_TEST_ADDITIONAL_FLAGS),
    "ABSL_CLANG_CL_FLAGS":
        MSVC_BIG_WARNING_FLAGS + MSVC_DEFINES,
    "ABSL_CLANG_CL_TEST_FLAGS":
        MSVC_BIG_WARNING_FLAGS + MSVC_DEFINES + ABSL_LLVM_TEST_ADDITIONAL_FLAGS,
    "ABSL_MSVC_FLAGS":
        MSVC_BIG_WARNING_FLAGS + MSVC_WARNING_FLAGS + MSVC_DEFINES,
    "ABSL_MSVC_TEST_FLAGS":
        MSVC_BIG_WARNING_FLAGS + MSVC_WARNING_FLAGS + MSVC_DEFINES + [
            "/wd4018",  # signed/unsigned mismatch
            "/wd4101",  # unreferenced local variable
            "/wd4503",  # decorated name length exceeded, name was truncated
            "/wd4996",  # use of deprecated symbol
            "/DNOMINMAX",  # disable the min() and max() macros from <windows.h>
        ],
    "ABSL_MSVC_LINKOPTS": [
        # Object file doesn't export any previously undefined symbols
        "-ignore:4221",
    ],
}



# Copyright 2025 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
from open_spiel.python.algorithms import ismcts
from open_spiel.python.algorithms import mcts
import pyspiel


NOLIMIT_GAME_STRING = (
    "universal_poker(betting=nolimit,bettingAbstraction=fullgame,blind=50 100,"
    "firstPlayer=1 1,numBoardCards=0 3 1 1,numHoleCards=2,numPlayers=2,"
    "numRanks=13,numRounds=2,numSuits=4,stack=20000 20000)"
)


class IsmctsTest(parameterized.TestCase):

  def test_action_candidates_selection(self):
    ismcts_bot = ismcts.ISMCTSBot(
        game=None,
        uct_c=1.0,
        evaluator=None,
        max_simulations=10,
    )

    # Test that the tie tolerance is respected.
    node = ismcts.ISMCTSNode()
    node.child_info = {
        0: ismcts.ChildInfo(visits=1, return_sum=7.0, prior=1.0),
        1: ismcts.ChildInfo(
            visits=1, return_sum=7.0 - ismcts.TIE_TOLERANCE / 2.0, prior=1.0
        ),
        2: ismcts.ChildInfo(
            visits=1, return_sum=7.0 - ismcts.TIE_TOLERANCE * 2.0, prior=1.0
        ),
    }
    node.total_visits = 4
    self.assertAlmostEqual(
        ismcts_bot._action_value(node, node.child_info[0]).item(), 8.0
    )
    candidates = ismcts_bot._select_candidate_actions(node)
    self.assertLen(candidates, 2)

    # Child 0 and 1 are selected because they are within the tie tolerance.
    self.assertIn(0, candidates)
    self.assertIn(1, candidates)

    # Child 2 is not selected because it is outside the tie tolerance.
    self.assertNotIn(2, candidates)

  def play_game(self, game: pyspiel.Game, ismcts_bot: ismcts.ISMCTSBot):
    state = game.new_initial_state()
    while not state.is_terminal():
      if state.is_chance_node():
        outcomes = state.chance_outcomes()
        action_list, prob_list = zip(*outcomes)
        action = np.random.choice(action_list, p=prob_list)
        state.apply_action(action)
      else:
        action = ismcts_bot.step(state)
        state.apply_action(action)

  def test_play_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    ismcts_bot = ismcts.ISMCTSBot(
        game=game,
        uct_c=4.0,
        evaluator=mcts.RandomRolloutEvaluator(),
        max_simulations=10,
    )
    self.play_game(game, ismcts_bot)

  @absltest.skip("Skipping. This one does not work.")
  def test_play_universal_poker(self):
    if "universal_poker" in pyspiel.registered_names():
      game = pyspiel.load_game(NOLIMIT_GAME_STRING)
      ismcts_bot = ismcts.ISMCTSBot(
          game=game,
          uct_c=4.0,
          evaluator=mcts.RandomRolloutEvaluator(),
          max_simulations=10,
      )
      self.play_game(game, ismcts_bot)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the min-max algorithm with alpha-beta pruning.

Solves perfect play for deterministic, 2-players, perfect-information 0-sum
games.

See for example https://en.wikipedia.org/wiki/Alpha-beta_pruning
"""

import pyspiel


def _alpha_beta(state, depth, alpha, beta, value_function,
                maximizing_player_id):
  """An alpha-beta algorithm.

  Implements a min-max algorithm with alpha-beta pruning.
  See for example https://en.wikipedia.org/wiki/Alpha-beta_pruning

  Arguments:
    state: The current state node of the game.
    depth: The maximum depth for the min/max search.
    alpha: best value that the MAX player can guarantee (if the value is <= than
      alpha, the MAX player will avoid it).
    beta: the best value that the MIN currently can guarantee (if the value is
      >= than beta, the MIN player will avoid it).
    value_function: An optional function mapping a Spiel `State` to a numerical
      value, to be used as the value of the maximizing player for a node when we
      reach `maximum_depth` and the node is not terminal.
    maximizing_player_id: The id of the MAX player. The other player is assumed
      to be MIN.

  Returns:
    A tuple of the optimal value of the sub-game starting in state
    (given alpha/beta) and the move that achieved it.

  Raises:
    NotImplementedError: If we reach the maximum depth. Given we have no value
      function for a non-terminal node, we cannot break early.
  """
  if state.is_terminal():
    return state.player_return(maximizing_player_id), None

  if depth == 0 and value_function is None:
    raise NotImplementedError(
        "We assume we can walk the full depth of the tree. "
        "Try increasing the maximum_depth or provide a value_function.")
  if depth == 0:
    return value_function(state), None

  player = state.current_player()
  best_action = -1
  if player == maximizing_player_id:
    value = -float("inf")
    for action in state.legal_actions():
      child_state = state.clone()
      child_state.apply_action(action)
      child_value, _ = _alpha_beta(child_state, depth - 1, alpha, beta,
                                   value_function, maximizing_player_id)
      if child_value > value:
        value = child_value
        best_action = action
      alpha = max(alpha, value)
      if alpha >= beta:
        break  # beta cut-off
    return value, best_action
  else:
    value = float("inf")
    for action in state.legal_actions():
      child_state = state.clone()
      child_state.apply_action(action)
      child_value, _ = _alpha_beta(child_state, depth - 1, alpha, beta,
                                   value_function, maximizing_player_id)
      if child_value < value:
        value = child_value
        best_action = action
      beta = min(beta, value)
      if alpha >= beta:
        break  # alpha cut-off
    return value, best_action


def alpha_beta_search(game,
                      state=None,
                      value_function=None,
                      maximum_depth=30,
                      maximizing_player_id=None):
  """Solves deterministic, 2-players, perfect-information 0-sum game.

  For small games only! Please use keyword arguments for optional arguments.

  Arguments:
    game: The game to analyze, as returned by `load_game`.
    state: The state to run from, as returned by `game.new_initial_state()`.  If
      none is specified, then the initial state is assumed.
    value_function: An optional function mapping a Spiel `State` to a numerical
      value, to be used as the value of the maximizing player for a node when we
      reach `maximum_depth` and the node is not terminal.
    maximum_depth: The maximum depth to search over. When this depth is reached,
      an exception will be raised.
    maximizing_player_id: The id of the MAX player. The other player is assumed
      to be MIN. The default (None) will suppose the player at the root to be
      the MAX player.

  Returns:
    A tuple containing the value of the game for the maximizing player when
    both player play optimally, and the action that achieves this value.
  """
  game_info = game.get_type()

  if game.num_players() != 2:
    raise ValueError("Game must be a 2-player game")
  if game_info.chance_mode != pyspiel.GameType.ChanceMode.DETERMINISTIC:
    raise ValueError("The game must be a Deterministic one, not {}".format(
        game.chance_mode))
  if game_info.information != pyspiel.GameType.Information.PERFECT_INFORMATION:
    raise ValueError(
        "The game must be a perfect information one, not {}".format(
            game.information))
  if game_info.dynamics != pyspiel.GameType.Dynamics.SEQUENTIAL:
    raise ValueError("The game must be turn-based, not {}".format(
        game.dynamics))
  if game_info.utility != pyspiel.GameType.Utility.ZERO_SUM:
    raise ValueError("The game must be 0-sum, not {}".format(game.utility))

  if state is None:
    state = game.new_initial_state()
  if maximizing_player_id is None:
    maximizing_player_id = state.current_player()
  return _alpha_beta(
      state.clone(),
      maximum_depth,
      alpha=-float("inf"),
      beta=float("inf"),
      value_function=value_function,
      maximizing_player_id=maximizing_player_id)


def expectiminimax(state, depth, value_function, maximizing_player_id):
  """Runs expectiminimax until the specified depth.

  See https://en.wikipedia.org/wiki/Expectiminimax for details.

  Arguments:
    state: The state to start the search from.
    depth: The depth of the search (not counting chance nodes).
    value_function: A value function, taking in a state and returning a value,
      in terms of the maximizing_player_id.
    maximizing_player_id: The player running the search (current player at root
      of the search tree).

  Returns:
    A tuple (value, best_action) representing the value to the maximizing player
    and the best action that achieves that value. None is returned as the best
    action at chance nodes, the depth limit, and terminals.
  """
  if state.is_terminal():
    return state.player_return(maximizing_player_id), None

  if depth == 0:
    return value_function(state), None

  if state.is_chance_node():
    value = 0
    for outcome, prob in state.chance_outcomes():
      child_state = state.clone()
      child_state.apply_action(outcome)
      child_value, _ = expectiminimax(child_state, depth, value_function,
                                      maximizing_player_id)
      value += prob * child_value
    return value, None
  elif state.current_player() == maximizing_player_id:
    value = -float("inf")
    for action in state.legal_actions():
      child_state = state.clone()
      child_state.apply_action(action)
      child_value, _ = expectiminimax(child_state, depth - 1, value_function,
                                      maximizing_player_id)
      if child_value > value:
        value = child_value
        best_action = action
    return value, best_action
  else:
    value = float("inf")
    for action in state.legal_actions():
      child_state = state.clone()
      child_state.apply_action(action)
      child_value, _ = expectiminimax(child_state, depth - 1, value_function,
                                      maximizing_player_id)
      if child_value < value:
        value = child_value
        best_action = action
    return value, best_action
